# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re
import requests
from json import dumps

from .NotifyBase import NotifyBase
from ..common import NotifyType
from ..utils import validate_regex
from ..AppriseLocale import gettext_lazy as _

# Used to break path apart into list of targets
TARGET_LIST_DELIM = re.compile(r'[ \t\r\n,\\/]+')


# Priorities
class ParsePlatformDevice:
    # All Devices
    ALL = 'all'

    # Apple IOS (APNS)
    IOS = 'ios'

    # Android/Firebase (FCM)
    ANDROID = 'android'


PARSE_PLATFORM_DEVICES = (
    ParsePlatformDevice.ALL,
    ParsePlatformDevice.IOS,
    ParsePlatformDevice.ANDROID,
)


class NotifyParsePlatform(NotifyBase):
    """
    A wrapper for Parse Platform Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'Parse Platform'

    # The services URL
    service_url = ' https://parseplatform.org/'

    # insecure notifications (using http)
    protocol = 'parsep'

    # Secure notifications (using https)
    secure_protocol = 'parseps'

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_parseplatform'

    # Define object templates
    templates = (
        '{schema}://{app_id}:{master_key}@{host}',
        '{schema}://{app_id}:{master_key}@{host}:{port}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'host': {
            'name': _('Hostname'),
            'type': 'string',
            'required': True,
        },
        'port': {
            'name': _('Port'),
            'type': 'int',
            'min': 1,
            'max': 65535,
        },
        'app_id': {
            'name': _('App ID'),
            'type': 'string',
            'private': True,
            'required': True,
        },
        'master_key': {
            'name': _('Master Key'),
            'type': 'string',
            'private': True,
            'required': True,
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'device': {
            'name': _('Device'),
            'type': 'choice:string',
            'values': PARSE_PLATFORM_DEVICES,
            'default': ParsePlatformDevice.ALL,
        },
        'app_id': {
            'alias_of': 'app_id',
        },
        'master_key': {
            'alias_of': 'master_key',
        },
    })

    def __init__(self, app_id, master_key, device=None, **kwargs):
        """
        Initialize Parse Platform Object
        """
        super().__init__(**kwargs)

        self.fullpath = kwargs.get('fullpath')
        if not isinstance(self.fullpath, str):
            self.fullpath = '/'

        # Application ID
        self.application_id = validate_regex(app_id)
        if not self.application_id:
            msg = 'An invalid Parse Platform Application ID ' \
                  '({}) was specified.'.format(app_id)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Master Key
        self.master_key = validate_regex(master_key)
        if not self.master_key:
            msg = 'An invalid Parse Platform Master Key ' \
                  '({}) was specified.'.format(master_key)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Initialize Devices Array
        self.devices = []

        if device:
            self.device = device.lower()
            if device not in PARSE_PLATFORM_DEVICES:
                msg = 'An invalid Parse Platform device ' \
                      '({}) was specified.'.format(device)
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.device = self.template_args['device']['default']

        if self.device == ParsePlatformDevice.ALL:
            self.devices = [d for d in PARSE_PLATFORM_DEVICES
                            if d != ParsePlatformDevice.ALL]
        else:
            # Store our device
            self.devices.append(device)

        return

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        Perform Parse Platform Notification
        """

        # Prepare our headers:
        headers = {
            'User-Agent': self.app_id,
            'Content-Type': 'application/json',
            'X-Parse-Application-Id': self.application_id,
            'X-Parse-Master-Key': self.master_key,
        }

        # prepare our payload
        payload = {
            'where': {
                'deviceType': {
                    '$in': self.devices,
                }
            },
            'data': {
                'title': title,
                'alert': body,
            }
        }

        # Set our schema
        schema = 'https' if self.secure else 'http'

        # Our Notification URL
        url = '%s://%s' % (schema, self.host)
        if isinstance(self.port, int):
            url += ':%d' % self.port

        url += self.fullpath.rstrip('/') + '/parse/push/'

        self.logger.debug('Parse Platform POST URL: %s (cert_verify=%r)' % (
            url, self.verify_certificate,
        ))
        self.logger.debug('Parse Platform Payload: %s' % str(payload))

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyParsePlatform.\
                    http_response_code_lookup(r.status_code)

                self.logger.warning(
                    'Failed to send Parse Platform notification: '
                    '{}{}error={}.'.format(
                        status_str,
                        ', ' if status_str else '',
                        r.status_code))

                self.logger.debug('Response Details:\r\n{}'.format(r.content))

                # Return; we're done
                return False

            else:
                self.logger.info('Sent Parse Platform notification.')

        except requests.RequestException as e:
            self.logger.warning(
                'A Connection error occured sending Parse Platform '
                'notification to %s.' % self.host)
            self.logger.debug('Socket Exception: %s' % str(e))

            # Return; we're done
            return False

        return True

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any arguments set
        params = {
            'device': self.device,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        default_port = 443 if self.secure else 80

        return \
            '{schema}://{app_id}:{master_key}@' \
            '{hostname}{port}{fullpath}/?{params}'.format(
                schema=self.secure_protocol if self.secure else self.protocol,
                app_id=self.pprint(self.application_id, privacy, safe=''),
                master_key=self.pprint(self.master_key, privacy, safe=''),
                hostname=NotifyParsePlatform.quote(self.host, safe=''),
                port='' if self.port is None or self.port == default_port
                     else ':{}'.format(self.port),
                fullpath=NotifyParsePlatform.quote(self.fullpath, safe='/'),
                params=NotifyParsePlatform.urlencode(params))

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to substantiate this object.

        """
        results = NotifyBase.parse_url(url)

        if not results:
            # We're done early as we couldn't load the results
            return results

        # App ID is retrieved from the user
        results['app_id'] = NotifyParsePlatform.unquote(results['user'])

        # Master Key is retrieved from the password
        results['master_key'] = \
            NotifyParsePlatform.unquote(results['password'])

        # Device support override
        if 'device' in results['qsd'] and len(results['qsd']['device']):
            results['device'] = results['qsd']['device']

        # Allow app_id attribute over-ride
        if 'app_id' in results['qsd'] and len(results['qsd']['app_id']):
            results['app_id'] = results['qsd']['app_id']

        # Allow master_key attribute over-ride
        if 'master_key' in results['qsd'] \
                and len(results['qsd']['master_key']):
            results['master_key'] = results['qsd']['master_key']

        return results



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import requests
from json import loads
from unittest import mock
from apprise.plugins.NotifyD7Networks import NotifyD7Networks
from helpers import AppriseURLTester
from apprise import Apprise
from apprise import NotifyType

# Disable logging for a cleaner testing output
import logging
logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    ('d7sms://', {
        # We failed to identify any valid authentication
        'instance': TypeError,
    }),
    ('d7sms://:@/', {
        # We failed to identify any valid authentication
        'instance': TypeError,
    }),
    ('d7sms://token@{}/{}/{}'.format('1' * 9, '2' * 15, 'a' * 13), {
        # No valid targets to notify
        'instance': NotifyD7Networks,
        # Since there are no targets specified we expect a False return on
        # send()
        'notify_response': False,
    }),
    ('d7sms://token1@{}?batch=yes'.format('3' * 14), {
        # valid number
        'instance': NotifyD7Networks,
        # Our expected url(privacy=True) startswith() response:
        'privacy_url': 'd7sms://t...1@',
    }),
    ('d7sms://token:colon2@{}?batch=yes'.format('3' * 14), {
        # valid number - token containing a colon
        'instance': NotifyD7Networks,
        # Our expected url(privacy=True) startswith() response:
        'privacy_url': 'd7sms://t...2@',
    }),
    ('d7sms://:token3@{}?batch=yes'.format('3' * 14), {
        # valid number - token starting wit a colon
        'instance': NotifyD7Networks,
        # Our expected url(privacy=True) startswith() response:
        'privacy_url': 'd7sms://:...3@',
    }),
    ('d7sms://{}?token=token6'.format('3' * 14), {
        # valid number - token starting wit a colon
        'instance': NotifyD7Networks,
        # Our expected url(privacy=True) startswith() response:
        'privacy_url': 'd7sms://t...6@',
    }),
    ('d7sms://token4@{}?unicode=no'.format('3' * 14), {
        # valid number - test unicode
        'instance': NotifyD7Networks,
        # Our expected url(privacy=True) startswith() response:
        'privacy_url': 'd7sms://t...4@',
    }),
    ('d7sms://token8@{}/{}/?unicode=yes'.format('3' * 14, '4' * 14), {
        # valid number - test unicode
        'instance': NotifyD7Networks,
        # Our expected url(privacy=True) startswith() response:
        'privacy_url': 'd7sms://t...8@',
    }),
    ('d7sms://token@{}?batch=yes&to={}'.format('3' * 14, '6' * 14), {
        # valid number
        'instance': NotifyD7Networks,
    }),
    ('d7sms://token@{}?batch=yes&from=apprise'.format('3' * 14), {
        # valid number, utilizing the optional from= variable
        'instance': NotifyD7Networks,
    }),
    ('d7sms://token@{}?batch=yes&source=apprise'.format('3' * 14), {
        # valid number, utilizing the optional source= variable (same as from)
        'instance': NotifyD7Networks,
    }),
    ('d7sms://token@{}?batch=no'.format('3' * 14), {
        # valid number - no batch
        'instance': NotifyD7Networks,
    }),
    ('d7sms://token@{}'.format('3' * 14), {
        'instance': NotifyD7Networks,
        # throw a bizzare code forcing us to fail to look it up
        'response': False,
        'requests_response_code': 999,
    }),
    ('d7sms://token@{}'.format('3' * 14), {
        'instance': NotifyD7Networks,
        # Throws a series of connection and transfer exceptions when this flag
        # is set and tests that we gracfully handle them
        'test_requests_exceptions': True,
    }),
)


def test_plugin_d7networks_urls():
    """
    NotifyD7Networks() Apprise URLs

    """

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch('requests.post')
def test_plugin_d7networks_edge_cases(mock_post):
    """
    NotifyD7Networks() Edge Cases tests

    """

    # Prepare Mock
    request = mock.Mock()
    request.content = '{}'
    request.status_code = requests.codes.ok
    mock_post.return_value = request

    # Initializations
    aobj = Apprise()
    assert aobj.add('d7sms://Token@15551231234/15551231236')

    body = "test message"

    # Send our notification
    assert aobj.notify(
        body=body, title='title', notify_type=NotifyType.INFO)

    # Not set to batch, so we send 2 different messages
    assert mock_post.call_count == 2
    assert mock_post.call_args_list[0][0][0] == \
        'https://api.d7networks.com/messages/v1/send'
    assert mock_post.call_args_list[1][0][0] == \
        'https://api.d7networks.com/messages/v1/send'

    # our first post
    data = loads(mock_post.call_args_list[0][1]['data'])
    assert len(data['messages']) == 1
    message = data['messages'][0]
    assert len(message['recipients']) == 1
    assert message['content'] == 'title\r\ntest message'
    assert message['data_coding'] == 'auto'

    # our second post
    data = loads(mock_post.call_args_list[1][1]['data'])
    assert len(data['messages']) == 1
    message = data['messages'][0]
    assert len(message['recipients']) == 1
    assert message['content'] == 'title\r\ntest message'
    assert message['data_coding'] == 'auto'

    #
    # Do a batch test now
    #

    mock_post.reset_mock()

    # Initializations
    aobj = Apprise()
    assert aobj.add('d7sms://Token@15551231234/15551231236?batch=yes')

    body = "test message"

    # Send our notification
    assert aobj.notify(
        body=body, title='title', notify_type=NotifyType.INFO)

    # All notifications go through in a batch
    assert mock_post.call_count == 1
    assert mock_post.call_args_list[0][0][0] == \
        'https://api.d7networks.com/messages/v1/send'

    data = loads(mock_post.call_args_list[0][1]['data'])
    assert len(data['messages']) == 1
    message = data['messages'][0]
    # All of our phone numbers were added here
    assert len(message['recipients']) == 2
    assert '15551231234' in message['recipients']
    assert '15551231236' in message['recipients']
    assert message['content'] == 'title\r\ntest message'
    assert message['data_coding'] == 'auto'



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# API Reference: https://smsmanager.cz/api/http#send

# To use this service you will need a SMS Manager account
# You will need credits (new accounts start with a few)
#     https://smsmanager.cz
#    1. Sign up and get test credit
#    2. Generate an API key in web administration.

import requests
from .NotifyBase import NotifyBase
from ..common import NotifyType
from ..utils import is_phone_no
from ..utils import parse_phone_no
from ..utils import parse_bool
from ..utils import validate_regex
from ..AppriseLocale import gettext_lazy as _


class SMSManagerGateway(object):
    """
    The different gateway values
    """
    HIGH = "high"
    ECONOMY = "economy"
    LOW = "low"
    DIRECT = "direct"


# Used for verification purposes
SMS_MANAGER_GATEWAYS = (
    SMSManagerGateway.HIGH,
    SMSManagerGateway.ECONOMY,
    SMSManagerGateway.LOW,
    SMSManagerGateway.DIRECT,
)


class NotifySMSManager(NotifyBase):
    """
    A wrapper for SMS Manager Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'SMS Manager'

    # The services URL
    service_url = 'https://smsmanager.cz'

    # All notification requests are secure
    secure_protocol = ('smsmgr', 'smsmanager',)

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_sms_manager'

    # SMS Manager uses the http protocol with JSON requests
    notify_url = 'https://http-api.smsmanager.cz/Send'

    # The maximum amount of texts that can go out in one batch
    default_batch_size = 4000

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        '{schema}://{apikey}@{targets}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'apikey': {
            'name': _('API Key'),
            'type': 'string',
            'private': True,
            'required': True,
        },
        'target_phone': {
            'name': _('Target Phone No'),
            'type': 'string',
            'prefix': '+',
            'regex': (r'^[0-9\s)(+-]+$', 'i'),
            'map_to': 'targets',
        },
        'targets': {
            'name': _('Targets'),
            'type': 'list:string',
            'required': True,
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'key': {
            'alias_of': 'apikey',
        },
        'to': {
            'alias_of': 'targets',
        },
        'from': {
            'name': _('From Phone No'),
            'type': 'string',
            'regex': (r'^\+?[0-9\s)(+-]+$', 'i'),
            'map_to': 'sender',
        },
        'sender': {
            'alias_of': 'from',
        },
        'gateway': {
            'name': _('Gateway'),
            'type': 'choice:string',
            'values': SMS_MANAGER_GATEWAYS,
            'default': SMS_MANAGER_GATEWAYS[0],
        },
        'batch': {
            'name': _('Batch Mode'),
            'type': 'bool',
            'default': False,
        },
    })

    def __init__(self, apikey=None, sender=None, targets=None, batch=None,
                 gateway=None, **kwargs):
        """
        Initialize SMS Manager Object
        """
        super(NotifySMSManager, self).__init__(**kwargs)

        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = 'An invalid API Key ({}) was specified.'.format(apikey)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Setup our gateway
        self.gateway = self.template_args['gateway']['default'] \
            if not isinstance(gateway, str) else gateway.lower()
        if self.gateway not in SMS_MANAGER_GATEWAYS:
            msg = 'The Gateway specified ({}) is invalid.'.format(gateway)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Define whether or not we should operate in a batch mode
        self.batch = self.template_args['batch']['default'] \
            if batch is None else bool(batch)

        # Maximum 11 characters and must be approved by administrators of site
        self.sender = sender[0:11] if isinstance(sender, str) else None

        # Parse our targets
        self.targets = list()

        for target in parse_phone_no(targets):
            # Parse each phone number we found
            # It is documented that numbers with a length of 9 characters are
            # supplemented by "420".
            result = is_phone_no(target, min_len=9)
            if result:
                # Carry forward '+' if defined, otherwise do not...
                self.targets.append(
                    ('+' + result['full'])
                    if target.lstrip()[0] == '+' else result['full'])
                continue

            self.logger.warning(
                'Dropped invalid phone # ({}) specified.'.format(target),
            )

        return

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        Perform SMS Manager Notification
        """

        if not self.targets:
            # We have nothing to notify
            self.logger.warning('There are no SMS Manager targets to notify')
            return False

        # error tracking (used for function return)
        has_error = False

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        # Prepare our headers
        headers = {
            'User-Agent': self.app_id,
        }

        # Prepare our targets
        targets = list(self.targets) if batch_size == 1 else \
            [self.targets[index:index + batch_size]
             for index in range(0, len(self.targets), batch_size)]

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our payload
            # Note: Payload is assembled inside of our while-loop due to
            #       mock testing issues (payload singleton isn't persistent
            #       when performing follow up checks on the params object.
            payload = {
                'apikey': self.apikey,
                'gateway': self.gateway,
                # The number gets populated in the loop below
                'number': None,
                'message': body,
            }

            if self.sender:
                # Sender is ony set if specified
                payload['sender'] = self.sender

            # Printable target details
            if isinstance(target, list):
                p_target = '{} targets'.format(len(target))

                # Prepare our target numbers
                payload['number'] = ';'.join(target)

            else:
                p_target = target
                # Prepare our target numbers
                payload['number'] = target

            # Some Debug Logging
            self.logger.debug(
                'SMS Manager POST URL: {} (cert_verify={})'.format(
                    self.notify_url, self.verify_certificate))
            self.logger.debug('SMS Manager Payload: {}' .format(payload))

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.get(
                    self.notify_url,
                    params=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = \
                        NotifyBase.http_response_code_lookup(r.status_code)

                    # set up our status code to use
                    status_code = r.status_code

                    self.logger.warning(
                        'Failed to send SMS Manager notification to {}: '
                        '{}{}error={}.'.format(
                            p_target,
                            status_str,
                            ', ' if status_str else '',
                            status_code))

                    self.logger.debug(
                        'Response Details:\r\n{}'.format(r.content))

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        'Sent SMS Manager notification to {}.'.format(
                            p_target))

            except requests.RequestException as e:
                self.logger.warning(
                    'A Connection error occurred sending SMS Manager: to %s ',
                    p_target)
                self.logger.debug('Socket Exception: %s' % str(e))

                # Mark our failure
                has_error = True
                continue

        return not has_error

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            'batch': 'yes' if self.batch else 'no',
            'gateway': self.gateway,
        }

        if self.sender:
            # Set our sender if it was set
            params['sender'] = self.sender

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return '{schema}://{apikey}@{targets}' \
            '?{params}'.format(
                schema=self.secure_protocol[0],
                apikey=self.pprint(self.apikey, privacy, safe=''),
                targets='/'.join([
                    NotifySMSManager.quote('{}'.format(x), safe='+')
                    for x in self.targets]),
                params=NotifySMSManager.urlencode(params))

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """

        #
        # Factor batch into calculation
        #
        # Note: Groups always require a separate request (and can not be
        # included in batch calculations)
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + \
                (1 if targets % batch_size else 0)

        return targets

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our API Key
        results['apikey'] = NotifySMSManager.unquote(results['user'])

        # Store our targets
        results['targets'] = [
            *NotifySMSManager.parse_phone_no(results['host']),
            *NotifySMSManager.split_path(results['fullpath'])]

        # The 'from' makes it easier to use yaml configuration
        if 'from' in results['qsd'] and len(results['qsd']['from']):
            results['sender'] = \
                NotifySMSManager.unquote(results['qsd']['from'])

        elif 'sender' in results['qsd'] and len(results['qsd']['sender']):
            # Support sender= value as well to align with SMS Manager API
            results['sender'] = \
                NotifySMSManager.unquote(results['qsd']['sender'])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if 'to' in results['qsd'] and len(results['qsd']['to']):
            results['targets'] += \
                NotifySMSManager.parse_phone_no(results['qsd']['to'])

        if 'key' in results['qsd'] and len(results['qsd']['key']):
            results['apikey'] = \
                NotifySMSManager.unquote(results['qsd']['key'])

        # Get Batch Mode Flag
        results['batch'] = \
            parse_bool(results['qsd'].get(
                'batch', NotifySMSManager.template_args['batch']['default']))

        # Define our gateway
        if 'gateway' in results['qsd'] and len(results['qsd']['gateway']):
            results['gateway'] = \
                NotifySMSManager.unquote(results['qsd']['gateway'])

        return results



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import pytest
from os.path import getsize
from os.path import join
from os.path import dirname
from inspect import cleandoc
from apprise.AttachmentManager import AttachmentManager
from apprise.AppriseAttachment import AppriseAttachment
from apprise.AppriseAsset import AppriseAsset
from apprise.attachment.AttachBase import AttachBase
from apprise.common import ContentLocation

# Disable logging for a cleaner testing output
import logging
logging.disable(logging.CRITICAL)

TEST_VAR_DIR = join(dirname(__file__), 'var')

# Grant access to our Attachment Manager Singleton
A_MGR = AttachmentManager()


def test_apprise_attachment():
    """
    API: AppriseAttachment basic testing

    """

    # Create ourselves an attachment object
    aa = AppriseAttachment()

    # There are no attachents loaded
    assert len(aa) == 0

    # Object can be directly checked as a boolean; response is False
    # when there are no entries loaded
    assert not aa

    # An attachment object using a custom Apprise Asset object
    # Set a cache expiry of 5 minutes (300 seconds)
    aa = AppriseAttachment(asset=AppriseAsset(), cache=300)

    # still no attachments added
    assert len(aa) == 0

    # Add a file by it's path
    path = join(TEST_VAR_DIR, 'apprise-test.gif')
    assert aa.add(path)

    # There is now 1 attachment
    assert len(aa) == 1

    # our attachment took on our cache value
    assert aa[0].cache == 300

    # we can test the object as a boolean and get a value of True now
    assert aa

    # Add another entry already in it's AttachBase format
    response = AppriseAttachment.instantiate(path, cache=True)
    assert isinstance(response, AttachBase)
    assert aa.add(response, asset=AppriseAsset())

    # There is now 2 attachments
    assert len(aa) == 2

    # Cache is initialized to True
    assert aa[1].cache is True

    # Reset our object
    aa = AppriseAttachment()

    # We can add by lists as well in a variety of formats
    attachments = (
        path,
        'file://{}?name=newfilename.gif?cache=120'.format(path),
        AppriseAttachment.instantiate(
            'file://{}?name=anotherfilename.gif'.format(path), cache=100),
    )

    # Add them
    assert aa.add(attachments, cache=False)

    # There is now 3 attachments
    assert len(aa) == 3

    # Take on our fixed cache value of False.
    # The last entry will have our set value of 100
    assert aa[0].cache is False
    # Even though we set a value of 120, we take on the value of False because
    # it was forced on the instantiate call
    assert aa[1].cache is False
    assert aa[2].cache == 100

    # We can pop the last element off of the list as well
    attachment = aa.pop()
    assert isinstance(attachment, AttachBase)
    # we can test of the attachment is valid using a boolean check:
    assert attachment
    assert len(aa) == 2
    assert attachment.path == path
    assert attachment.name == 'anotherfilename.gif'
    assert attachment.mimetype == 'image/gif'

    # elements can also be directly indexed
    assert isinstance(aa[0], AttachBase)
    assert isinstance(aa[1], AttachBase)

    with pytest.raises(IndexError):
        aa[2]

    # We can iterate over attachments too:
    for count, a in enumerate(aa):
        assert isinstance(a, AttachBase)

        # we'll never iterate more then the number of entries in our object
        assert count < len(aa)

    # Get the file-size of our image
    expected_size = getsize(path) * len(aa)

    # verify that's what we get as a result
    assert aa.size() == expected_size

    # Attachments can also be loaded during the instantiation of the
    # AppriseAttachment object
    aa = AppriseAttachment(attachments)

    # There is now 3 attachments
    assert len(aa) == 3

    # Reset our object
    aa.clear()
    assert len(aa) == 0
    assert not aa

    assert aa.add(AppriseAttachment.instantiate(
        'file://{}?name=andanother.png&cache=Yes'.format(path)))
    assert aa.add(AppriseAttachment.instantiate(
        'file://{}?name=andanother.png&cache=No'.format(path)))
    AppriseAttachment.instantiate(
        'file://{}?name=andanother.png&cache=600'.format(path))
    assert aa.add(AppriseAttachment.instantiate(
        'file://{}?name=andanother.png&cache=600'.format(path)))

    assert len(aa) == 3
    assert aa[0].cache is True
    assert aa[1].cache is False
    assert aa[2].cache == 600

    # Negative cache are not allowed
    assert not aa.add(AppriseAttachment.instantiate(
        'file://{}?name=andanother.png&cache=-600'.format(path)))

    # Invalid cache value
    assert not aa.add(AppriseAttachment.instantiate(
        'file://{}?name=andanother.png'.format(path), cache='invalid'))

    # No length change
    assert len(aa) == 3

    # Reset our object
    aa.clear()

    # Garbage in produces garbage out
    assert aa.add(None) is False
    assert aa.add(object()) is False
    assert aa.add(42) is False

    # length remains unchanged
    assert len(aa) == 0

    # We can add by lists as well in a variety of formats
    attachments = (
        None,
        object(),
        42,
        'garbage://',
    )

    # Add our attachments
    assert aa.add(attachments) is False

    # length remains unchanged
    assert len(aa) == 0

    # if instantiating attachments from the class, it will throw a TypeError
    # if attachments couldn't be loaded
    with pytest.raises(TypeError):
        AppriseAttachment('garbage://')

    # Load our other attachment types
    aa = AppriseAttachment(location=ContentLocation.LOCAL)

    # Hosted type won't allow us to import files
    aa = AppriseAttachment(location=ContentLocation.HOSTED)
    assert len(aa) == 0

    # Add our attachments defined a the head of this function
    aa.add(attachments)

    # Our length is still zero because we can't import files in
    # a hosted environment
    assert len(aa) == 0

    # Inaccessible type prevents the adding of new stuff
    aa = AppriseAttachment(location=ContentLocation.INACCESSIBLE)
    assert len(aa) == 0

    # Add our attachments defined a the head of this function
    aa.add(attachments)

    # Our length is still zero
    assert len(aa) == 0

    with pytest.raises(TypeError):
        # Invalid location specified
        AppriseAttachment(location="invalid")

    # test cases when file simply doesn't exist
    aa = AppriseAttachment('file://non-existant-file.png')
    # Our length is still 1
    assert len(aa) == 1
    # Our object will still return a True
    assert aa

    # However our indexed entry will not
    assert not aa[0]

    # length will return 0
    assert len(aa[0]) == 0

    # Total length will also return 0
    assert aa.size() == 0


def test_apprise_attachment_instantiate():
    """
    API: AppriseAttachment.instantiate()

    """
    assert AppriseAttachment.instantiate(
        'file://?', suppress_exceptions=True) is None

    assert AppriseAttachment.instantiate(
        'invalid://?', suppress_exceptions=True) is None

    class BadAttachType(AttachBase):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            # We fail whenever we're initialized
            raise TypeError()

    # Store our bad attachment type in our schema map
    A_MGR['bad'] = BadAttachType

    with pytest.raises(TypeError):
        AppriseAttachment.instantiate(
            'bad://path', suppress_exceptions=False)

    # Same call but exceptions suppressed
    assert AppriseAttachment.instantiate(
        'bad://path', suppress_exceptions=True) is None


def test_attachment_matrix_dynamic_importing(tmpdir):
    """
    API: Apprise() Attachment Matrix Importing

    """

    # Make our new path valid
    suite = tmpdir.mkdir("apprise_attach_test_suite")
    suite.join("__init__.py").write('')

    module_name = 'badattach'

    # Create a base area to work within
    base = suite.mkdir(module_name)
    base.join("__init__.py").write('')

    # Test no app_id
    base.join('AttachBadFile1.py').write(cleandoc(
        """
        class AttachBadFile1:
            pass
        """))

    # No class of the same name
    base.join('AttachBadFile2.py').write(cleandoc(
        """
        class BadClassName:
            pass
        """))

    # Exception thrown
    base.join('AttachBadFile3.py').write("""raise ImportError()""")

    # Utilizes a schema:// already occupied (as string)
    base.join('AttachGoober.py').write(cleandoc(
        """
        from apprise import AttachBase
        class AttachGoober(AttachBase):
            # This class tests the fact we have a new class name, but we're
            # trying to over-ride items previously used

            # The default simple (insecure) protocol
            protocol = 'http'

            # The default secure protocol
            secure_protocol = 'https'
        """))

    # Utilizes a schema:// already occupied (as tuple)
    base.join('AttachBugger.py').write(cleandoc(
        """
        from apprise import AttachBase
        class AttachBugger(AttachBase):
            # This class tests the fact we have a new class name, but we're
            # trying to over-ride items previously used

            # The default simple (insecure) protocol
            protocol = ('http', 'bugger-test' )

            # The default secure protocol
            secure_protocol = ('https', 'bugger-tests')
        """))

    A_MGR.load_modules(path=str(base), name=module_name)



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import pickle
from apprise import Apprise, AppriseAsset, AppriseLocale

# Disable logging for a cleaner testing output
import logging
logging.disable(logging.CRITICAL)

# Ensure we don't create .pyc files for these tests
sys.dont_write_bytecode = True


def test_apprise_pickle_asset(tmpdir):
    """pickle: AppriseAsset
    """
    asset = AppriseAsset()
    serialized = pickle.dumps(asset)
    new_asset = pickle.loads(serialized)

    # iterate over some keys to verify they're still the same:
    keys = (
        'app_id', 'app_desc', 'app_url', 'html_notify_map',
        'ascii_notify_map', 'default_html_color', 'default_extension',
        'theme', 'image_url_mask', 'image_url_logo', 'image_path_mask',
        'body_format', 'async_mode', 'interpret_escapes', 'encoding',
        'secure_logging', '_recursion',
    )

    for key in keys:
        assert getattr(asset, key) == getattr(new_asset, key)


def test_apprise_pickle_locale(tmpdir):
    """pickle: AppriseLocale
    """
    _locale = AppriseLocale.AppriseLocale()
    serialized = pickle.dumps(_locale)
    new_locale = pickle.loads(serialized)

    assert _locale.lang == new_locale.lang

    # Ensure internal functions still call in new object
    new_locale.detect_language()


def test_apprise_pickle_core(tmpdir):
    """pickle: Apprise
    """
    asset = AppriseAsset(app_id="default")
    apobj = Apprise(asset=asset)

    # Create a custom asset so we can verify it gets correctly serialized
    xml_asset = AppriseAsset(app_id="xml")

    # Store our Entries
    apobj.add("json://localhost")
    apobj.add("xml://localhost", asset=xml_asset)
    apobj.add("form://localhost")
    apobj.add("mailto://user:pass@localhost", tag="email")
    serialized = pickle.dumps(apobj)

    # Unserialize our object
    new_apobj = pickle.loads(serialized)

    # Verify that it loaded our URLs back
    assert len(new_apobj) == 4

    # Our assets were kept (note the XML altered entry)
    assert apobj[0].app_id == "default"
    assert apobj[1].app_id == "xml"
    assert apobj[2].app_id == "default"
    assert apobj[3].app_id == "default"

    # Our tag was kept
    assert "email" in apobj[3].tags



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import asyncio
import concurrent.futures
import re
import sys
import pytest
import requests
from inspect import cleandoc
from unittest import mock

from os.path import dirname
from os.path import join

from apprise import Apprise
from apprise import AppriseAsset
from apprise import AppriseAttachment
from apprise import NotifyBase
from apprise import NotifyType
from apprise import NotifyFormat
from apprise import NotifyImageSize
from apprise import __version__
from apprise import URLBase
from apprise import PrivacyMode
from apprise.AppriseLocale import LazyTranslation
from apprise.AppriseLocale import gettext_lazy as _

from apprise.NotificationManager import NotificationManager
from apprise.utils import parse_list
from helpers import OuterEventLoop
import inspect

# Disable logging for a cleaner testing output
import logging
logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = join(dirname(__file__), 'var')

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()


def test_apprise_object():
    """
    API: Apprise() object

    """
    def do_notify(server, *args, **kwargs):
        return server.notify(*args, **kwargs)

    apprise_test(do_notify)


def test_apprise_async():
    """
    API: Apprise() object asynchronous methods

    """
    with OuterEventLoop() as loop:
        def do_notify(server, *args, **kwargs):
            return loop.run_until_complete(
                server.async_notify(*args, **kwargs))

        apprise_test(do_notify)


def apprise_test(do_notify):
    a = Apprise()

    # no items
    assert len(a) == 0

    # Apprise object can also be directly tested with 'if' keyword
    # No entries results in a False response
    assert not a

    # Create an Asset object
    asset = AppriseAsset(theme='default')

    # We can load the device using our asset
    a = Apprise(asset=asset)

    # We can load our servers up front as well
    servers = [
        'faast://abcdefghijklmnop-abcdefg',
        'kodi://kodi.server.local',
    ]

    a = Apprise(servers=servers)

    # 2 servers loaded
    assert len(a) == 2

    # Apprise object can also be directly tested with 'if' keyword
    # At least one entry results in a True response
    assert a

    # We can retrieve our URLs this way:
    assert len(a.urls()) == 2

    # We can add another server
    assert a.add('mmosts://mattermost.server.local/'
                 '3ccdd113474722377935511fc85d3dd4') is True
    assert len(a) == 3

    # Try adding nothing but delimiters
    assert a.add(',, ,, , , , ,') is False

    # The number of servers added doesn't change
    assert len(a) == 3

    # We can pop an object off of our stack by it's indexed value:
    obj = a.pop(0)
    assert isinstance(obj, NotifyBase) is True
    assert len(a) == 2

    # We can retrieve elements from our list too by reference:
    assert isinstance(a[0].url(), str) is True

    # We can iterate over our list too:
    count = 0
    for o in a:
        assert isinstance(o.url(), str) is True
        count += 1
    # verify that we did indeed iterate over each element
    assert len(a) == count

    # We can empty our set
    a.clear()
    assert len(a) == 0

    # An invalid schema
    assert a.add('this is not a parseable url at all') is False
    assert len(a) == 0

    # An unsupported schema
    assert a.add(
        'invalid://we.just.do.not.support.this.plugin.type') is False
    assert len(a) == 0

    # A poorly formatted URL
    assert a.add('json://user:@@@:bad?no.good') is False
    assert len(a) == 0

    # Add a server with our asset we created earlier
    assert a.add('mmosts://mattermost.server.local/'
                 '3ccdd113474722377935511fc85d3dd4', asset=asset) is True

    # Clear our server listings again
    a.clear()

    # No servers to notify
    assert do_notify(a, title="my title", body="my body") is False

    class BadNotification(NotifyBase):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            # We fail whenever we're initialized
            raise TypeError()

        def url(self, **kwargs):
            # Support URL
            return ''

        @staticmethod
        def parse_url(url, *args, **kwargs):
            # always parseable
            return NotifyBase.parse_url(url, verify_host=False)

    class GoodNotification(NotifyBase):
        def __init__(self, **kwargs):
            super().__init__(
                notify_format=NotifyFormat.HTML, **kwargs)

        def url(self, **kwargs):
            # Support URL
            return ''

        def send(self, **kwargs):
            # Pretend everything is okay
            return True

        @staticmethod
        def parse_url(url, *args, **kwargs):
            # always parseable
            return NotifyBase.parse_url(url, verify_host=False)

    # Store our bad notification in our schema map
    N_MGR['bad'] = BadNotification

    # Store our good notification in our schema map
    N_MGR['good'] = GoodNotification

    # Just to explain what is happening here, we would have parsed the
    # url properly but failed when we went to go and create an instance
    # of it.
    assert a.add('bad://localhost') is False
    assert len(a) == 0

    # We'll fail because we've got nothing to notify
    assert do_notify(
        a, title="my title", body="my body") is False

    # Clear our server listings again
    a.clear()

    assert a.add('good://localhost') is True
    assert len(a) == 1

    # Bad Notification Type is still allowed as it is presumed the user
    # know's what their doing
    assert do_notify(
        a, title="my title", body="my body", notify_type='bad') is True

    # No Title/Body combo's
    assert do_notify(a, title=None, body=None) is False
    assert do_notify(a, title='', body=None) is False
    assert do_notify(a, title=None, body='') is False

    assert do_notify(a, title=5, body=b'bytes') is False
    assert do_notify(a, title=b"bytes", body=10) is False
    assert do_notify(a, title=object(), body=b'bytes') is False
    assert do_notify(a, title=b"bytes", body=object()) is False

    # A Body must be present
    assert do_notify(a, title='present', body=None) is False

    # Other combinations work fine
    assert do_notify(a, title=None, body='present') is True
    assert do_notify(a, title="present", body="present") is True

    # Send Attachment with success
    attach = join(TEST_VAR_DIR, 'apprise-test.gif')
    assert do_notify(
        a, body='body', title='test', notify_type=NotifyType.INFO,
        attach=attach) is True

    # Send the attachment as an AppriseAttachment object
    assert do_notify(
        a, body='body', title='test', notify_type=NotifyType.INFO,
        attach=AppriseAttachment(attach)) is True

    # test a invalid attachment
    assert do_notify(
        a, body='body', title='test', notify_type=NotifyType.INFO,
        attach='invalid://') is False

    # Repeat the same tests above...
    # however do it by directly accessing the object; this grants the similar
    # results:
    assert do_notify(
        a[0], body='body', title='test', notify_type=NotifyType.INFO,
        attach=attach) is True

    # Send the attachment as an AppriseAttachment object
    assert do_notify(
        a[0], body='body', title='test', notify_type=NotifyType.INFO,
        attach=AppriseAttachment(attach)) is True

    # test a invalid attachment
    assert do_notify(
        a[0], body='body', title='test', notify_type=NotifyType.INFO,
        attach='invalid://') is False

    class ThrowNotification(NotifyBase):
        def notify(self, **kwargs):
            # Pretend everything is okay
            raise TypeError()

        async def async_notify(self, **kwargs):
            # Pretend everything is okay (async)
            raise TypeError()

        def url(self, **kwargs):
            # Support URL
            return ''

    class RuntimeNotification(NotifyBase):
        def notify(self, **kwargs):
            # Pretend everything is okay
            raise RuntimeError()

        async def async_notify(self, **kwargs):
            # Pretend everything is okay (async)
            raise TypeError()

        def url(self, **kwargs):
            # Support URL
            return ''

    class FailNotification(NotifyBase):

        def notify(self, **kwargs):
            # Pretend everything is okay
            return False

        async def async_notify(self, **kwargs):
            # Pretend everything is okay (async)
            raise TypeError()

        def url(self, **kwargs):
            # Support URL
            return ''

    # Store our bad notification in our schema map
    N_MGR['throw'] = ThrowNotification

    # Store our good notification in our schema map
    N_MGR['fail'] = FailNotification

    # Store our good notification in our schema map
    N_MGR['runtime'] = RuntimeNotification

    for async_mode in (True, False):
        # Create an Asset object
        asset = AppriseAsset(theme='default', async_mode=async_mode)

        # We can load the device using our asset
        a = Apprise(asset=asset)

        assert a.add('runtime://localhost') is True
        assert a.add('throw://localhost') is True
        assert a.add('fail://localhost') is True
        assert len(a) == 3

        # Test when our notify both throws an exception and or just
        # simply returns False
        assert do_notify(a, title="present", body="present") is False

    # Create a Notification that throws an unexected exception
    class ThrowInstantiateNotification(NotifyBase):
        def __init__(self, **kwargs):
            # Pretend everything is okay
            raise TypeError()

        def url(self, **kwargs):
            # Support URL
            return ''

    N_MGR.unload_modules()
    N_MGR['throw'] = ThrowInstantiateNotification

    # Store our good notification in our schema map
    N_MGR['good'] = GoodNotification

    # Reset our object
    a.clear()
    assert len(a) == 0

    # Test our socket details
    # rto = Socket Read Timeout
    # cto = Socket Connect Timeout
    plugin = a.instantiate('good://localhost?rto=5.1&cto=10')
    assert isinstance(plugin, NotifyBase)
    assert plugin.socket_connect_timeout == 10.0
    assert plugin.socket_read_timeout == 5.1

    plugin = a.instantiate('good://localhost?rto=invalid&cto=invalid')
    assert isinstance(plugin, NotifyBase)
    assert plugin.socket_connect_timeout == URLBase.socket_connect_timeout
    assert plugin.socket_read_timeout == URLBase.socket_read_timeout

    # Reset our object
    a.clear()
    assert len(a) == 0

    # Instantiate a bad object
    plugin = a.instantiate(object, tag="bad_object")
    assert plugin is None

    # Instantiate a good object
    plugin = a.instantiate('good://localhost', tag="good")
    assert isinstance(plugin, NotifyBase)

    # Test simple tagging inside of the object
    assert "good" in plugin
    assert "bad" not in plugin

    # the in (__contains__ override) is based on or'ed content; so although
    # 'bad' isn't tagged as being in the plugin, 'good' is, so the return
    # value of this is True
    assert ["bad", "good"] in plugin
    assert set(["bad", "good"]) in plugin
    assert ("bad", "good") in plugin

    # We an add already substatiated instances into our Apprise object
    a.add(plugin)
    assert len(a) == 1

    # We can add entries as a list too (to add more then one)
    a.add([plugin, plugin, plugin])
    assert len(a) == 4

    # Reset our object again
    a.clear()
    with pytest.raises(TypeError):
        a.instantiate('throw://localhost', suppress_exceptions=False)

    assert len(a) == 0

    assert a.instantiate(
        'throw://localhost', suppress_exceptions=True) is None
    assert len(a) == 0

    #
    # We rince and repeat the same tests as above, however we do them
    # using the dict version
    #

    # Reset our object
    a.clear()
    assert len(a) == 0

    # Instantiate a good object
    plugin = a.instantiate({
        'schema': 'good',
        'host': 'localhost'}, tag="good")
    assert isinstance(plugin, NotifyBase)

    # Test simple tagging inside of the object
    assert "good" in plugin
    assert "bad" not in plugin

    # the in (__contains__ override) is based on or'ed content; so although
    # 'bad' isn't tagged as being in the plugin, 'good' is, so the return
    # value of this is True
    assert ["bad", "good"] in plugin
    assert set(["bad", "good"]) in plugin
    assert ("bad", "good") in plugin

    # We an add already substatiated instances into our Apprise object
    a.add(plugin)
    assert len(a) == 1

    # We can add entries as a list too (to add more then one)
    a.add([plugin, plugin, plugin])
    assert len(a) == 4

    # Reset our object again
    a.clear()
    with pytest.raises(TypeError):
        a.instantiate({
            'schema': 'throw',
            'host': 'localhost'}, suppress_exceptions=False)

    assert len(a) == 0

    assert a.instantiate({
        'schema': 'throw',
        'host': 'localhost'}, suppress_exceptions=True) is None
    assert len(a) == 0


def test_apprise_pretty_print(tmpdir):
    """
    API: Apprise() Pretty Print tests

    """
    # Privacy Print
    # PrivacyMode.Secret always returns the same thing to avoid guessing
    assert URLBase.pprint(
        None, privacy=True, mode=PrivacyMode.Secret) == '****'
    assert URLBase.pprint(
        42, privacy=True, mode=PrivacyMode.Secret) == '****'
    assert URLBase.pprint(
        object, privacy=True, mode=PrivacyMode.Secret) == '****'
    assert URLBase.pprint(
        "", privacy=True, mode=PrivacyMode.Secret) == '****'
    assert URLBase.pprint(
        "a", privacy=True, mode=PrivacyMode.Secret) == '****'
    assert URLBase.pprint(
        "ab", privacy=True, mode=PrivacyMode.Secret) == '****'
    assert URLBase.pprint(
        "abcdefghijk", privacy=True, mode=PrivacyMode.Secret) == '****'

    # PrivacyMode.Outer
    assert URLBase.pprint(
        None, privacy=True, mode=PrivacyMode.Outer) == ''
    assert URLBase.pprint(
        42, privacy=True, mode=PrivacyMode.Outer) == ''
    assert URLBase.pprint(
        object, privacy=True, mode=PrivacyMode.Outer) == ''
    assert URLBase.pprint(
        "", privacy=True, mode=PrivacyMode.Outer) == ''
    assert URLBase.pprint(
        "a", privacy=True, mode=PrivacyMode.Outer) == 'a...a'
    assert URLBase.pprint(
        "ab", privacy=True, mode=PrivacyMode.Outer) == 'a...b'
    assert URLBase.pprint(
        "abcdefghijk", privacy=True, mode=PrivacyMode.Outer) == 'a...k'

    # PrivacyMode.Tail
    assert URLBase.pprint(
        None, privacy=True, mode=PrivacyMode.Tail) == ''
    assert URLBase.pprint(
        42, privacy=True, mode=PrivacyMode.Tail) == ''
    assert URLBase.pprint(
        object, privacy=True, mode=PrivacyMode.Tail) == ''
    assert URLBase.pprint(
        "", privacy=True, mode=PrivacyMode.Tail) == ''
    assert URLBase.pprint(
        "a", privacy=True, mode=PrivacyMode.Tail) == '...a'
    assert URLBase.pprint(
        "ab", privacy=True, mode=PrivacyMode.Tail) == '...ab'
    assert URLBase.pprint(
        "abcdefghijk", privacy=True, mode=PrivacyMode.Tail) == '...hijk'

    # Quoting settings
    assert URLBase.pprint(" ", privacy=False, safe='') == '%20'
    assert URLBase.pprint(" ", privacy=False, quote=False, safe='') == ' '


@mock.patch('requests.get')
@mock.patch('requests.post')
def test_apprise_tagging(mock_post, mock_get):
    """
    API: Apprise() object tagging functionality

    """
    def do_notify(server, *args, **kwargs):
        return server.notify(*args, **kwargs)

    apprise_tagging_test(mock_post, mock_get, do_notify)


@mock.patch('requests.get')
@mock.patch('requests.post')
def test_apprise_tagging_async(mock_post, mock_get):
    """
    API: Apprise() object tagging functionality asynchronous methods

    """
    with OuterEventLoop() as loop:
        def do_notify(server, *args, **kwargs):
            return loop.run_until_complete(
                server.async_notify(*args, **kwargs))

        apprise_tagging_test(mock_post, mock_get, do_notify)


def apprise_tagging_test(mock_post, mock_get, do_notify):
    # A request
    robj = mock.Mock()
    setattr(robj, 'raw', mock.Mock())
    # Allow raw.read() calls
    robj.raw.read.return_value = ''
    robj.text = ''
    robj.content = ''
    mock_get.return_value = robj
    mock_post.return_value = robj

    # Simulate a successful notification
    mock_get.return_value.status_code = requests.codes.ok
    mock_post.return_value.status_code = requests.codes.ok

    # Create our object
    a = Apprise()

    # An invalid addition can't add the tag
    assert a.add('averyinvalidschema://localhost', tag='uhoh') is False
    assert a.add({
        'schema': 'averyinvalidschema',
        'host': 'localhost'}, tag='uhoh') is False

    # Add entry and assign it to a tag called 'awesome'
    assert a.add('json://localhost/path1/', tag='awesome') is True
    assert a.add({
        'schema': 'json',
        'host': 'localhost',
        'fullpath': '/path1/'}, tag='awesome') is True

    # Add another notification and assign it to a tag called 'awesome'
    # and another tag called 'local'
    assert a.add('json://localhost/path2/', tag=['mmost', 'awesome']) is True

    # notify the awesome tag; this would notify both services behind the
    # scenes
    assert do_notify(
        a, title="my title", body="my body", tag='awesome') is True

    # notify all of the tags
    assert do_notify(
        a, title="my title", body="my body", tag=['awesome', 'mmost']) is True

    # When we query against our loaded notifications for a tag that simply
    # isn't assigned to anything, we return None.  None (different then False)
    # tells us that we litterally had nothing to query.  We didn't fail...
    # but we also didn't do anything...
    assert do_notify(
        a, title="my title", body="my body", tag='missing') is None

    # Now to test the ability to and and/or notifications
    a = Apprise()

    # Add a tag by tuple
    assert a.add('json://localhost/tagA/', tag=("TagA", )) is True
    # Add 2 tags by string
    assert a.add('json://localhost/tagAB/', tag="TagA, TagB") is True
    # Add a tag using a set
    assert a.add('json://localhost/tagB/', tag=set(["TagB"])) is True
    # Add a tag by string (again)
    assert a.add('json://localhost/tagC/', tag="TagC") is True
    # Add 2 tags using a list
    assert a.add('json://localhost/tagCD/', tag=["TagC", "TagD"]) is True
    # Add a tag by string (again)
    assert a.add('json://localhost/tagD/', tag="TagD") is True
    # add a tag set by set (again)
    assert a.add('json://localhost/tagCDE/',
                 tag=set(["TagC", "TagD", "TagE"])) is True

    # Expression: TagC and TagD
    # Matches the following only:
    #   - json://localhost/tagCD/
    #   - json://localhost/tagCDE/
    assert do_notify(
        a, title="my title", body="my body", tag=[('TagC', 'TagD')]) is True

    # Expression: (TagY and TagZ) or TagX
    # Matches nothing, None is returned in this case
    assert do_notify(
        a, title="my title", body="my body",
        tag=[('TagY', 'TagZ'), 'TagX']) is None

    # Expression: (TagY and TagZ) or TagA
    # Matches the following only:
    #   - json://localhost/tagAB/
    assert do_notify(
        a, title="my title", body="my body",
        tag=[('TagY', 'TagZ'), 'TagA']) is True

    # Expression: (TagE and TagD) or TagB
    # Matches the following only:
    #   - json://localhost/tagCDE/
    #   - json://localhost/tagAB/
    #   - json://localhost/tagB/
    assert do_notify(
        a, title="my title", body="my body",
        tag=[('TagE', 'TagD'), 'TagB']) is True

    # Garbage Entries in tag field just get stripped out. the below
    # is the same as notifying no tags at all. Since we have not added
    # any entries that do not have tags (that we can match against)
    # we fail.  None is returned as a way of letting us know that we
    # had Notifications to notify, but since none of them matched our tag
    # none were notified.
    assert do_notify(
        a, title="my title", body="my body",
        tag=[(object, ), ]) is None


def test_apprise_schemas(tmpdir):
    """
    API: Apprise().schema() tests

    """
    # Clear loaded modules
    a = Apprise()

    # no items
    assert len(a) == 0

    class TextNotification(NotifyBase):
        # set our default notification format
        notify_format = NotifyFormat.TEXT

        # Garbage Protocol Entries
        protocol = None

        secure_protocol = (None, object)

    class HtmlNotification(NotifyBase):

        protocol = ('html', 'htm')

        secure_protocol = ('htmls', 'htms')

    class MarkDownNotification(NotifyBase):

        protocol = 'markdown'

        secure_protocol = 'markdowns'

    schemas = URLBase.schemas(TextNotification)
    assert isinstance(schemas, set) is True
    # We didn't define a protocol or secure protocol
    assert len(schemas) == 0

    # Store our notifications into our schema map
    N_MGR['text'] = TextNotification
    N_MGR['html'] = HtmlNotification
    N_MGR['markdown'] = MarkDownNotification

    schemas = URLBase.schemas(TextNotification)
    assert isinstance(schemas, set) is True
    # We didn't define a protocol or secure protocol one
    # but one got assigned in he above N_MGR call
    assert len(schemas) == 1
    assert 'text' in schemas

    schemas = URLBase.schemas(HtmlNotification)
    assert isinstance(schemas, set) is True
    assert len(schemas) == 4
    assert 'html' in schemas
    assert 'htm' in schemas
    assert 'htmls' in schemas
    assert 'htms' in schemas

    # Invalid entries do not disrupt schema calls
    for garbage in (object(), None, 42):
        schemas = URLBase.schemas(garbage)
        assert isinstance(schemas, set) is True
        assert len(schemas) == 0


def test_apprise_urlbase_object():
    """
    API: Apprise() URLBase object testing

    """
    results = URLBase.parse_url('https://localhost/path/?cto=3.0&verify=no')
    assert results.get('user') is None
    assert results.get('password') is None
    assert results.get('path') == '/path/'
    assert results.get('secure') is True
    assert results.get('verify') is False
    base = URLBase(**results)
    assert base.request_timeout == (3.0, 4.0)
    assert base.request_auth is None
    assert base.request_url == 'https://localhost/path/'
    assert base.url().startswith('https://localhost/')

    results = URLBase.parse_url(
        'http://user:pass@localhost:34/path/here?rto=3.0&verify=yes')
    assert results.get('user') == 'user'
    assert results.get('password') == 'pass'
    assert results.get('fullpath') == '/path/here'
    assert results.get('secure') is False
    assert results.get('verify') is True
    base = URLBase(**results)
    assert base.request_timeout == (4.0, 3.0)
    assert base.request_auth == ('user', 'pass')
    assert base.request_url == 'http://localhost:34/path/here'
    assert base.url().startswith('http://user:pass@localhost:34/path/here')

    results = URLBase.parse_url('http://user@127.0.0.1/path/')
    assert results.get('user') == 'user'
    assert results.get('password') is None
    assert results.get('fullpath') == '/path/'
    assert results.get('secure') is False
    assert results.get('verify') is True
    base = URLBase(**results)
    assert base.request_timeout == (4.0, 4.0)
    assert base.request_auth == ('user', None)
    assert base.request_url == 'http://127.0.0.1/path/'
    assert base.url().startswith('http://user@127.0.0.1/path/')


def test_apprise_notify_formats(tmpdir):
    """
    API: Apprise() Input Formats tests

    """
    # Need to set async_mode=False to call notify() instead of async_notify().
    asset = AppriseAsset(async_mode=False)

    a = Apprise(asset=asset)

    # no items
    assert len(a) == 0

    class TextNotification(NotifyBase):
        # set our default notification format
        notify_format = NotifyFormat.TEXT

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        def notify(self, **kwargs):
            # Pretend everything is okay
            return True

        def url(self, **kwargs):
            # Support URL
            return ''

    class HtmlNotification(NotifyBase):

        # set our default notification format
        notify_format = NotifyFormat.HTML

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        def notify(self, **kwargs):
            # Pretend everything is okay
            return True

        def url(self, **kwargs):
            # Support URL
            return ''

    class MarkDownNotification(NotifyBase):

        # set our default notification format
        notify_format = NotifyFormat.MARKDOWN

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        def notify(self, **kwargs):
            # Pretend everything is okay
            return True

        def url(self, **kwargs):
            # Support URL
            return ''

    # Store our notifications into our schema map
    N_MGR['text'] = TextNotification
    N_MGR['html'] = HtmlNotification
    N_MGR['markdown'] = MarkDownNotification

    # Test Markdown; the above calls the markdown because our good://
    # defined plugin above was defined to default to HTML which triggers
    # a markdown to take place if the body_format specified on the notify
    # call
    assert a.add('html://localhost') is True
    assert a.add('html://another.server') is True
    assert a.add('html://and.another') is True
    assert a.add('text://localhost') is True
    assert a.add('text://another.server') is True
    assert a.add('text://and.another') is True
    assert a.add('markdown://localhost') is True
    assert a.add('markdown://another.server') is True
    assert a.add('markdown://and.another') is True

    assert len(a) == 9

    assert a.notify(
        title="markdown", body="## Testing Markdown",
        body_format=NotifyFormat.MARKDOWN) is True

    assert a.notify(
        title="text", body="Testing Text",
        body_format=NotifyFormat.TEXT) is True

    assert a.notify(
        title="html", body="<b>HTML</b>",
        body_format=NotifyFormat.HTML) is True


def test_apprise_asset(tmpdir):
    """
    API: AppriseAsset() object

    """
    a = AppriseAsset(theme='light')
    # Default theme
    assert a.theme == 'light'

    # Invalid kw handling
    with pytest.raises(AttributeError):
        AppriseAsset(invalid_kw='value')

    a = AppriseAsset(
        theme='dark',
        image_path_mask='/{THEME}/{TYPE}-{XY}{EXTENSION}',
        image_url_mask='http://localhost/{THEME}/{TYPE}-{XY}{EXTENSION}',
    )

    a.default_html_color = '#abcabc'

    assert a.color('invalid', tuple) == (171, 202, 188)
    assert a.color(NotifyType.INFO, tuple) == (58, 163, 227)

    assert a.color('invalid', int) == 11258556
    assert a.color(NotifyType.INFO, int) == 3843043

    assert a.color('invalid', None) == '#abcabc'
    assert a.color(NotifyType.INFO, None) == '#3AA3E3'
    # None is the default
    assert a.color(NotifyType.INFO) == '#3AA3E3'

    # Invalid Type
    with pytest.raises(ValueError):
        # The exception we expect since dict is not supported
        a.color(NotifyType.INFO, dict)

    assert a.image_url(NotifyType.INFO, NotifyImageSize.XY_256) == \
        'http://localhost/dark/info-256x256.png'

    assert a.image_path(
        NotifyType.INFO,
        NotifyImageSize.XY_256,
        must_exist=False) == '/dark/info-256x256.png'

    # This path doesn't exist so image_raw will fail (since we just
    # randompyl picked it for testing)
    assert a.image_raw(NotifyType.INFO, NotifyImageSize.XY_256) is None

    assert a.image_path(
        NotifyType.INFO,
        NotifyImageSize.XY_256,
        must_exist=True) is None

    # Create a new object (with our default settings)
    a = AppriseAsset()

    # Our default configuration can access our file
    assert a.image_path(
        NotifyType.INFO,
        NotifyImageSize.XY_256,
        must_exist=True) is not None

    assert a.image_raw(NotifyType.INFO, NotifyImageSize.XY_256) is not None

    # Create a temporary directory
    sub = tmpdir.mkdir("great.theme")

    # Write a file
    sub.join("{0}-{1}.png".format(
        NotifyType.INFO,
        NotifyImageSize.XY_256,
    )).write("the content doesn't matter for testing.")

    # Create an asset that will reference our file we just created
    a = AppriseAsset(
        theme='great.theme',
        image_path_mask='%s/{THEME}/{TYPE}-{XY}.png' % dirname(sub.strpath),
    )

    # We'll be able to read file we just created
    assert a.image_raw(NotifyType.INFO, NotifyImageSize.XY_256) is not None

    # We can retrieve the filename at this point even with must_exist set
    # to True
    assert a.image_path(
        NotifyType.INFO,
        NotifyImageSize.XY_256,
        must_exist=True) is not None

    # Test case where we can't access the image file
    with mock.patch('builtins.open', side_effect=OSError()):
        assert a.image_raw(NotifyType.INFO, NotifyImageSize.XY_256) is None

    # Our content is retrivable again
    assert a.image_raw(NotifyType.INFO, NotifyImageSize.XY_256) is not None

    # Disable all image references
    a = AppriseAsset(image_path_mask=False, image_url_mask=False)

    # We always return none in these calls now
    assert a.image_raw(NotifyType.INFO, NotifyImageSize.XY_256) is None
    assert a.image_url(NotifyType.INFO, NotifyImageSize.XY_256) is None
    assert a.image_path(NotifyType.INFO, NotifyImageSize.XY_256,
                        must_exist=False) is None
    assert a.image_path(NotifyType.INFO, NotifyImageSize.XY_256,
                        must_exist=True) is None

    # Test our default extension out
    a = AppriseAsset(
        image_path_mask='/{THEME}/{TYPE}-{XY}{EXTENSION}',
        image_url_mask='http://localhost/{THEME}/{TYPE}-{XY}{EXTENSION}',
        default_extension='.jpeg',
    )
    assert a.image_path(
        NotifyType.INFO,
        NotifyImageSize.XY_256,
        must_exist=False) == '/default/info-256x256.jpeg'

    assert a.image_url(
        NotifyType.INFO,
        NotifyImageSize.XY_256) == \
        'http://localhost/default/info-256x256.jpeg'

    # extension support
    assert a.image_path(
        NotifyType.INFO,
        NotifyImageSize.XY_128,
        must_exist=False,
        extension='.ico') == '/default/info-128x128.ico'

    assert a.image_url(
        NotifyType.INFO,
        NotifyImageSize.XY_256,
        extension='.test') == \
        'http://localhost/default/info-256x256.test'


def test_apprise_disabled_plugins():
    """
    API: Apprise() Disabled Plugin States

    """
    # Ensure there are no other drives loaded
    N_MGR.unload_modules(disable_native=True)
    assert len(N_MGR) == 0

    class TestDisabled01Notification(NotifyBase):
        """
        This class is used to test a pre-disabled state
        """

        # Just flat out disable our service
        enabled = False

        # we'll use this as a key to make our service easier to find
        # in the next part of the testing
        service_name = 'na01'

        def url(self, **kwargs):
            # Support URL
            return ''

        def notify(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    N_MGR['na01'] = TestDisabled01Notification

    class TestDisabled02Notification(NotifyBase):
        """
        This class is used to test a post-disabled state
        """

        # we'll use this as a key to make our service easier to find
        # in the next part of the testing
        service_name = 'na02'

        def __init__(self, *args, **kwargs):
            super().__init__(**kwargs)

            # enable state changes **AFTER** we initialize
            self.enabled = False

        def url(self, **kwargs):
            # Support URL
            return ''

        def notify(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    N_MGR['na02'] = TestDisabled02Notification

    # Create our Apprise instance
    a = Apprise()

    result = a.details(lang='ca-en', show_disabled=True)
    assert isinstance(result, dict)
    assert 'schemas' in result
    assert len(result['schemas']) == 2

    # our na01 is disabled right from the get-go
    entry = next((x for x in result['schemas']
                  if x['service_name'] == 'na01'), None)
    assert entry is not None
    assert entry['enabled'] is False

    plugin = a.instantiate('na01://localhost')
    # Object is just flat out disabled... nothing is instatiated
    assert plugin is None

    # our na02 isn't however until it's initialized; as a result
    # it get's returned in our result set
    entry = next((x for x in result['schemas']
                  if x['service_name'] == 'na02'), None)
    assert entry is not None
    assert entry['enabled'] is True

    plugin = a.instantiate('na02://localhost')
    # Object isn't disabled until the __init__() call.  But this is still
    # enough to not instantiate the object:
    assert plugin is None

    # If we choose to filter our disabled, we can't unfortunately filter those
    # that go disabled after instantiation, but we do filter out any that are
    # already known to not be enabled:
    result = a.details(lang='ca-en', show_disabled=False)
    assert isinstance(result, dict)
    assert 'schemas' in result
    assert len(result['schemas']) == 1

    # We'll add a good notification to our list
    class TesEnabled01Notification(NotifyBase):
        """
        This class is just a simple enabled one
        """

        # we'll use this as a key to make our service easier to find
        # in the next part of the testing
        service_name = 'good'

        def url(self, **kwargs):
            # Support URL
            return ''

        def send(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    N_MGR['good'] = TesEnabled01Notification

    # The last thing we'll simulate is a case where the plugin is just
    # disabled at a later time long into it's life.  this is just to allow
    # administrators to stop the flow of their notifications for their own
    # given reasons.
    plugin = a.instantiate('good://localhost')
    assert isinstance(plugin, NotifyBase)

    # we'll toggle our state
    plugin.enabled = False

    # As a result, we can now no longer send a notification:
    assert plugin.notify("My Message") is False

    # As just a proof of how you can toggle the state back:
    plugin.enabled = True

    # our notifications will go okay now
    assert plugin.notify("My Message") is True

    # Restore our modules
    N_MGR.unload_modules()


def test_apprise_details():
    """
    API: Apprise() Details

    """
    # This is a made up class that is just used to verify
    class TestDetailNotification(NotifyBase):
        """
        This class is used to test various configurations supported
        """

        # Minimum requirements for a plugin to produce details
        service_name = 'Detail Testing'

        # The default simple (insecure) protocol (used by NotifyMail)
        protocol = 'details'

        # Set test_bool flag
        always_true = True
        always_false = False

        # Define object templates
        templates = (
            '{schema}://{host}',
            '{schema}://{host}:{port}',
            '{schema}://{user}@{host}:{port}',
            '{schema}://{user}:{pass}@{host}:{port}',
        )

        # Define our tokens; these are the minimum tokens required required to
        # be passed into this function (as arguments). The syntax appends any
        # previously defined in the base package and builds onto them
        template_tokens = dict(NotifyBase.template_tokens, **{
            'notype': {
                # name is a minimum requirement
                'name': _('no type'),
            },
            'regex_test01': {
                'name': _('RegexTest'),
                'type': 'string',
                'regex': r'^[A-Z0-9]$',
            },
            'regex_test02': {
                'name': _('RegexTest'),
                # Support regex options too
                'regex': (r'^[A-Z0-9]$', 'i'),
            },
            'regex_test03': {
                'name': _('RegexTest'),
                # Support regex option without a second option
                'regex': (r'^[A-Z0-9]$'),
            },
            'regex_test04': {
                # this entry would just end up getting removed
                'regex': None,
            },
            # List without delimiters (causes defaults to kick in)
            'mylistA': {
                'name': 'fruit',
                'type': 'list:string',
            },
            # A list with a delimiter list
            'mylistB': {
                'name': 'softdrinks',
                'type': 'list:string',
                'delim': ['|', '-'],
            },
        })

        template_args = dict(NotifyBase.template_args, **{
            # Test _exist_if logic
            'test_exists_if_01': {
                'name': 'Always False',
                'type': 'bool',
                # Provide a default
                'default': False,
                # Base the existance of this key/value entry on the lookup
                # of this class value at runtime. Hence:
                #     if not NotifyObject.always_false
                #         del this_entry
                #
                '_exists_if': 'always_false',
            },
            # Test _exist_if logic
            'test_exists_if_02': {
                'name': 'Always True',
                'type': 'bool',
                # Provide a default
                'default': False,
                # Base the existance of this key/value entry on the lookup
                # of this class value at runtime. Hence:
                #     if not NotifyObject.always_true
                #         del this_entry
                #
                '_exists_if': 'always_true',
            },
            # alias_of testing
            'test_alias_of': {
                'alias_of': 'mylistB',
                'delim': ('-', ' ')
            }
        })

        def url(self, **kwargs):
            # Support URL
            return ''

        def send(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    # Store our good detail notification in our schema map
    N_MGR['details'] = TestDetailNotification

    # This is a made up class that is just used to verify
    class TestReq01Notification(NotifyBase):
        """
        This class is used to test various requirement configurations
        """

        # Set some requirements
        requirements = {
            'packages_required': [
                'cryptography <= 3.4',
                'ultrasync',
            ],
            'packages_recommended': 'django',
        }

        def url(self, **kwargs):
            # Support URL
            return ''

        def send(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    N_MGR['req01'] = TestReq01Notification

    # This is a made up class that is just used to verify
    class TestReq02Notification(NotifyBase):
        """
        This class is used to test various requirement configurations
        """

        # Just not enabled at all
        enabled = False

        # Set some requirements
        requirements = {
            # None and/or [] is implied, but jsut to show that the code won't
            # crash if explicitly set this way:
            'packages_required': None,

            'packages_recommended': [
                'cryptography <= 3.4',
            ]
        }

        def url(self, **kwargs):
            # Support URL
            return ''

        def send(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    N_MGR['req02'] = TestReq02Notification

    # This is a made up class that is just used to verify
    class TestReq03Notification(NotifyBase):
        """
        This class is used to test various requirement configurations
        """

        # Set some requirements
        requirements = {
            # We can over-ride the default details assigned to our plugin if
            # specified
            'details': _('some specified requirement details'),

            # We can set a string value as well (it does not have to be a list)
            'packages_recommended': 'cryptography <= 3.4'
        }

        def url(self, **kwargs):
            # Support URL
            return ''

        def send(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    N_MGR['req03'] = TestReq03Notification

    # This is a made up class that is just used to verify
    class TestReq04Notification(NotifyBase):
        """
        This class is used to test a case where our requirements is fixed
        to a None
        """

        # This is the same as saying there are no requirements
        requirements = None

        def url(self, **kwargs):
            # Support URL
            return ''

        def send(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    N_MGR['req04'] = TestReq04Notification

    # This is a made up class that is just used to verify
    class TestReq05Notification(NotifyBase):
        """
        This class is used to test a case where only packages_recommended
        is identified
        """

        requirements = {
            # We can set a string value as well (it does not have to be a list)
            'packages_recommended': 'cryptography <= 3.4'
        }

        def url(self, **kwargs):
            # Support URL
            return ''

        def send(self, **kwargs):
            # Pretend everything is okay (so we don't break other tests)
            return True

    N_MGR['req05'] = TestReq05Notification

    # Create our Apprise instance
    a = Apprise()

    # Dictionary response
    result = a.details()
    assert isinstance(result, dict)

    # Test different variations of our call
    result = a.details(lang='ca-fr')
    assert isinstance(result, dict)
    for entry in result['schemas']:
        # Verify our key does not exist because we did not ask for it
        assert 'enabled' not in entry
        assert 'requirements' not in entry

    result = a.details(lang='us-en', show_requirements=True)
    assert isinstance(result, dict)
    for entry in result['schemas']:
        # Verify our key does not exist because we did not ask for it
        assert 'enabled' not in entry

        # Requirements are set for display
        assert 'requirements' in entry
        assert 'details' in entry['requirements']
        assert 'packages_required' in entry['requirements']
        assert 'packages_recommended' in entry['requirements']
        assert isinstance(entry['requirements']['details'], (
            str, LazyTranslation))
        assert isinstance(entry['requirements']['packages_required'], list)
        assert isinstance(entry['requirements']['packages_recommended'], list)

    result = a.details(lang='ca-en', show_disabled=True)
    assert isinstance(result, dict)
    for entry in result['schemas']:
        # Verify that our plugin state is available to us
        assert 'enabled' in entry
        assert isinstance(entry['enabled'], bool)

        # Verify our key does not exist because we did not ask for it
        assert 'requirements' not in entry

    result = a.details(
        lang='ca-fr', show_requirements=True, show_disabled=True)
    assert isinstance(result, dict)
    for entry in result['schemas']:
        # Plugin States are set for display
        assert 'enabled' in entry
        assert isinstance(entry['enabled'], bool)

        # Requirements are set for display
        assert 'requirements' in entry
        assert 'details' in entry['requirements']
        assert 'packages_required' in entry['requirements']
        assert 'packages_recommended' in entry['requirements']
        assert isinstance(entry['requirements']['details'], (
            str, LazyTranslation))
        assert isinstance(entry['requirements']['packages_required'], list)
        assert isinstance(entry['requirements']['packages_recommended'], list)


def test_apprise_details_plugin_verification():
    """
    API: Apprise() Details Plugin Verification

    """
    # Prepare our object
    a = Apprise()

    # Details object
    details = a.details()

    # Dictionary response
    assert isinstance(details, dict)

    # Details object with language defined:
    details = a.details(lang='en')

    # Dictionary response
    assert isinstance(details, dict)

    # Details object with unsupported language:
    details = a.details(lang='xx')

    # Dictionary response
    assert isinstance(details, dict)

    # Apprise version
    assert 'version' in details
    assert details.get('version') == __version__

    # Defined schemas identify each plugin
    assert 'schemas' in details
    assert isinstance(details.get('schemas'), list)

    # We have an entry per defined plugin
    assert 'asset' in details
    assert isinstance(details.get('asset'), dict)
    assert 'app_id' in details['asset']
    assert 'app_desc' in details['asset']
    assert 'default_extension' in details['asset']
    assert 'theme' in details['asset']
    assert 'image_path_mask' in details['asset']
    assert 'image_url_mask' in details['asset']
    assert 'image_url_logo' in details['asset']

    # Valid Type Regular Expression Checker
    # Case Sensitive and MUST match the following:
    is_valid_type_re = re.compile(
        r'((choice|list):)?(string|bool|int|float)')

    # match tokens found in templates so we can cross reference them back
    # to see if they have a matching argument
    template_token_re = re.compile(r'{([^}]+)}[^{]*?(?=$|{)')

    # Define acceptable map_to arguments that can be tied in with the
    # kwargs function definitions.
    valid_kwargs = set([
        # General Parameters
        'user', 'password', 'port', 'host', 'schema', 'fullpath',
        # NotifyBase parameters:
        'format', 'overflow', 'emojis',
        # URLBase parameters:
        'verify', 'cto', 'rto',
    ])

    # Valid Schema Entries:
    valid_schema_keys = (
        'name', 'private', 'required', 'type', 'values', 'min', 'max',
        'regex', 'default', 'list', 'delim', 'prefix', 'map_to', 'alias_of',
        'group',
    )
    for entry in details['schemas']:

        # Track the map_to entries (if specified); We need to make sure that
        # these properly map back
        map_to_entries = set()

        # Track the alias_of entries
        map_to_aliases = set()

        # A Service Name MUST be defined
        assert 'service_name' in entry
        assert isinstance(
            entry['service_name'], (str, LazyTranslation))

        # Acquire our protocols
        protocols = parse_list(
            entry['protocols'], entry['secure_protocols'])

        # At least one schema/protocol MUST be defined
        assert len(protocols) > 0

        # our details
        assert 'details' in entry
        assert isinstance(entry['details'], dict)

        # All schema details should include args
        for section in ['kwargs', 'args', 'tokens']:
            assert section in entry['details']
            assert isinstance(entry['details'][section], dict)

            for key, arg in entry['details'][section].items():
                # Validate keys (case-sensitive)
                assert len([k for k in arg.keys()
                            if k not in valid_schema_keys]) == 0

                # Test our argument
                assert isinstance(arg, dict)

                if 'alias_of' not in arg:
                    # Minimum requirement of an argument
                    assert 'name' in arg
                    assert isinstance(arg['name'], str)

                    assert 'type' in arg
                    assert isinstance(arg['type'], str)
                    assert is_valid_type_re.match(arg['type']) is not None

                    if 'min' in arg:
                        assert arg['type'].endswith('float') \
                            or arg['type'].endswith('int')
                        assert isinstance(arg['min'], (int, float))

                        if 'max' in arg:
                            # If a min and max was specified, at least check
                            # to confirm the min is less then the max
                            assert arg['min'] < arg['max']

                    if 'max' in arg:
                        assert arg['type'].endswith('float') \
                            or arg['type'].endswith('int')
                        assert isinstance(arg['max'], (int, float))

                    if 'private' in arg:
                        assert isinstance(arg['private'], bool)

                    if 'required' in arg:
                        assert isinstance(arg['required'], bool)

                    if 'prefix' in arg:
                        assert isinstance(arg['prefix'], str)
                        if section == 'kwargs':
                            # The only acceptable prefix types for kwargs
                            assert arg['prefix'] in (':', '+', '-')

                    else:
                        # kwargs requires that the 'prefix' is defined
                        assert section != 'kwargs'

                    if 'map_to' in arg:
                        # must be a string
                        assert isinstance(arg['map_to'], str)
                        # Track our map_to object
                        map_to_entries.add(arg['map_to'])

                    else:
                        map_to_entries.add(key)

                    # Some verification
                    if arg['type'].startswith('choice'):

                        # choice:bool is redundant and should be swapped to
                        # just bool
                        assert not arg['type'].endswith('bool')

                        # Choices require that a values list is provided
                        assert 'values' in arg
                        assert isinstance(arg['values'], (list, tuple))
                        assert len(arg['values']) > 0

                        # Test default
                        if 'default' in arg:
                            # if a default is provided on a choice object,
                            # it better be in the list of values
                            assert arg['default'] in arg['values']

                    if arg['type'].startswith('bool'):
                        # Boolean choices are less restrictive but require a
                        # default value
                        assert 'default' in arg
                        assert isinstance(arg['default'], bool)

                    if 'regex' in arg:
                        # Regex must ALWAYS be in the format (regex, option)
                        assert isinstance(arg['regex'], (tuple, list))
                        assert len(arg['regex']) == 2
                        assert isinstance(arg['regex'][0], str)
                        assert arg['regex'][1] is None or isinstance(
                            arg['regex'][1], str)

                        # Compile the regular expression to verify that it is
                        # valid
                        try:
                            re.compile(arg['regex'][0])
                        except:
                            assert '{} is an invalid regex'\
                                .format(arg['regex'][0])

                        # Regex should always start and/or end with ^/$
                        assert re.match(
                            r'^\^.+?$', arg['regex'][0]) is not None
                        assert re.match(
                            r'^.+?\$$', arg['regex'][0]) is not None

                    if arg['type'].startswith('list'):
                        # Delimiters MUST be defined
                        assert 'delim' in arg
                        assert isinstance(arg['delim'], (list, tuple))
                        assert len(arg['delim']) > 0

                else:  # alias_of is in the object
                    # Ensure we're not already in the tokens section
                    # The alias_of object has no value here
                    assert section != 'tokens'

                    # must be a string
                    assert isinstance(
                        arg['alias_of'], (str, list, tuple, set))

                    aliases = [arg['alias_of']] \
                        if isinstance(arg['alias_of'], str) \
                        else arg['alias_of']

                    for alias_of in aliases:
                        # Track our alias_of object
                        map_to_aliases.add(alias_of)

                        # We can't be an alias_of ourselves
                        if key == alias_of:
                            # This is acceptable as long as we exist in the
                            # tokens table because that is truely what we map
                            # back to
                            assert key in entry['details']['tokens']

                        else:
                            # Throw the problem into an assert tag for
                            # debugging purposes... the mapping is not
                            # acceptable
                            assert key != alias_of

                        # alias_of always references back to tokens
                        assert \
                            alias_of in entry['details']['tokens'] or \
                            alias_of in entry['details']['args']

                        # Find a list directive in our tokens
                        t_match = entry['details']['tokens']\
                            .get(alias_of, {})\
                            .get('type', '').startswith('list')

                        a_match = entry['details']['args']\
                            .get(alias_of, {})\
                            .get('type', '').startswith('list')

                        if not (t_match or a_match):
                            # Ensure the only token we have is the alias_of
                            # hence record should look like as example):
                            # {
                            #    'token': {
                            #      'alias_of': 'apitoken',
                            #    },
                            # }
                            #
                            # Or if it can represent more then one entry; in
                            # this case, one must define a name (to define
                            # grouping).
                            # {
                            #    'token': {
                            #      'name': 'Tokens',
                            #      'alias_of': ('apitoken', 'webtoken'),
                            #    },
                            # }
                            if isinstance(arg['alias_of'], str):
                                assert len(entry['details'][section][key]) == 1
                            else:  # is tuple,list, or set
                                assert len(entry['details'][section][key]) == 2
                                # Must have a name defined to define grouping
                                assert 'name' in entry['details'][section][key]

                        else:
                            # We're a list, we allow up to 2 variables
                            # Obviously we have the alias_of entry; that's why
                            # were at this part of the code.  But we can
                            # additionally provide a 'delim' over-ride.
                            assert len(entry['details'][section][key]) <= 2
                            if len(entry['details'][section][key]) == 2:
                                # Verify that it is in fact the 'delim' tag
                                assert 'delim' in \
                                    entry['details'][section][key]
                                # If we do have a delim value set, it must be
                                # of a list/set/tuple type
                                assert isinstance(
                                    entry['details'][section][key]['delim'],
                                    (tuple, set, list),
                                )

        spec = inspect.getfullargspec(
            N_MGR._schema_map[protocols[0]].__init__)

        function_args = \
            (set(parse_list(spec.varkw)) - set(['kwargs'])) \
            | (set(spec.args) - set(['self'])) | valid_kwargs

        # Iterate over our map_to_entries and make sure that everything
        # maps to a function argument
        for arg in map_to_entries:
            if arg not in function_args:
                # This print statement just makes the error easier to
                # troubleshoot
                raise AssertionError(
                    '{}.__init__() expects a {}=None entry according to '
                    'template configuration'
                    .format(
                        N_MGR._schema_map
                        [protocols[0]].__name__, arg))

        # Iterate over all of the function arguments and make sure that
        # it maps back to a key
        function_args -= valid_kwargs
        for arg in function_args:
            if arg not in map_to_entries:
                raise AssertionError(
                    '{}.__init__({}) found but not defined in the '
                    'template configuration'
                    .format(N_MGR._schema_map[protocols[0]].__name__, arg))

        # Iterate over our map_to_aliases and make sure they were defined in
        # either the as a token or arg
        for arg in map_to_aliases:
            assert arg in set(entry['details']['args'].keys()) \
                | set(entry['details']['tokens'].keys())

        # Template verification
        assert 'templates' in entry['details']
        assert isinstance(entry['details']['templates'], (set, tuple, list))

        # Iterate over our templates and parse our arguments
        for template in entry['details']['templates']:
            # Ensure we've properly opened and closed all of our tokens
            assert template.count('{') == template.count('}')

            expected_tokens = template.count('}')
            args = template_token_re.findall(template)
            assert expected_tokens == len(args)

            # Build a cross reference set of our current defined objects
            defined_tokens = set()
            for key, arg in entry['details']['tokens'].items():
                defined_tokens.add(key)
                if 'alias_of' in arg:
                    defined_tokens.add(arg['alias_of'])

            # We want to make sure all of our defined tokens have been
            # accounted for in at least one defined template
            for arg in args:
                assert arg in set(entry['details']['args'].keys()) \
                    | set(entry['details']['tokens'].keys())

                # The reverse of the above; make sure that each entry defined
                # in the template_tokens is accounted for in at least one of
                # the defined templates
                assert arg in defined_tokens


@mock.patch('requests.post')
@mock.patch('asyncio.gather', wraps=asyncio.gather)
@mock.patch('concurrent.futures.ThreadPoolExecutor',
            wraps=concurrent.futures.ThreadPoolExecutor)
def test_apprise_async_mode(mock_threadpool, mock_gather, mock_post, tmpdir):
    """
    API: Apprise() async_mode tests

    """
    mock_post.return_value.status_code = requests.codes.ok

    # Define some servers
    servers = [
        'xml://localhost',
        'json://localhost',
    ]

    # Default Async Mode is to be enabled
    asset = AppriseAsset()
    assert asset.async_mode is True

    # Load our asset
    a = Apprise(asset=asset)

    # add our servers
    a.add(servers=servers)

    # 2 servers loaded
    assert len(a) == 2

    # Our servers should carry this flag
    for server in a:
        assert server.asset.async_mode is True

    # Send Notifications Asyncronously
    assert a.notify("async") is True

    # Verify our thread pool was created
    assert mock_threadpool.call_count == 1
    mock_threadpool.reset_mock()

    # Provide an over-ride now
    asset = AppriseAsset(async_mode=False)
    assert asset.async_mode is False

    # Load our asset
    a = Apprise(asset=asset)

    # Verify our configuration kept
    assert a.asset.async_mode is False

    # add our servers
    a.add(servers=servers)

    # 2 servers loaded
    assert len(a) == 2

    # Our servers should carry this flag
    for server in a:
        assert server.asset.async_mode is False

    # Send Notifications Syncronously
    assert a.notify("sync") is True
    # Sequential send doesn't require a gather
    assert mock_gather.call_count == 0
    mock_gather.reset_mock()

    # another way of looking a our false set asset configuration
    assert a[0].asset.async_mode is False
    assert a[1].asset.async_mode is False

    # Adjust 1 of the servers async_mode settings
    a[0].asset.async_mode = True
    assert a[0].asset.async_mode is True

    # They all share the same object, so this gets toggled too
    assert a[1].asset.async_mode is True

    # We'll just change this one
    a[1].asset = AppriseAsset(async_mode=False)
    assert a[0].asset.async_mode is True
    assert a[1].asset.async_mode is False

    # Send 1 Notification Syncronously, the other Asyncronously
    assert a.notify("a mixed batch") is True

    # Verify we didn't use a thread pool for a single notification
    assert mock_threadpool.call_count == 0
    mock_threadpool.reset_mock()


def test_notify_matrix_dynamic_importing(tmpdir):
    """
    API: Apprise() Notify Matrix Importing

    """

    # Make our new path valid
    suite = tmpdir.mkdir("apprise_notify_test_suite")
    suite.join("__init__.py").write('')

    module_name = 'badnotify'

    # Update our path to point to our new test suite
    sys.path.insert(0, str(suite))

    # Create a base area to work within
    base = suite.mkdir(module_name)
    base.join("__init__.py").write('')

    # Test no app_id
    base.join('NotifyBadFile1.py').write(cleandoc(
        """
        class NotifyBadFile1:
            pass
        """))

    # No class of the same name
    base.join('NotifyBadFile2.py').write(cleandoc(
        """
        class BadClassName:
            pass
        """))

    # Exception thrown
    base.join('NotifyBadFile3.py').write("""raise ImportError()""")

    # Utilizes a schema:// already occupied (as string)
    base.join('NotifyGoober.py').write(cleandoc(
        """
        from apprise import NotifyBase
        class NotifyGoober(NotifyBase):
            # This class tests the fact we have a new class name, but we're
            # trying to over-ride items previously used

            # The default simple (insecure) protocol (used by NotifyMail)
            protocol = ('mailto', 'goober')

            # The default secure protocol (used by NotifyMail)
            secure_protocol = 'mailtos'

            @staticmethod
            def parse_url(url, *args, **kwargs):
                # always parseable
                return ConfigBase.parse_url(url, verify_host=False)
        """))

    # Utilizes a schema:// already occupied (as tuple)
    base.join('NotifyBugger.py').write(cleandoc(
        """
        from apprise import NotifyBase
        class NotifyBugger(NotifyBase):
            # This class tests the fact we have a new class name, but we're
            # trying to over-ride items previously used

            # The default simple (insecure) protocol (used by NotifyMail), the
            # other isn't
            protocol = ('mailto', 'bugger-test' )

            # The default secure protocol (used by NotifyMail), the other isn't
            secure_protocol = ('mailtos', ['garbage'])

            @staticmethod
            def parse_url(url, *args, **kwargs):
                # always parseable
                return ConfigBase.parse_url(url, verify_host=False)
            """))

    N_MGR.load_modules(path=str(base), name=module_name)



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from ..plugins.NotifyBase import NotifyBase
from ..NotificationManager import NotificationManager
from ..utils import URL_DETAILS_RE
from ..utils import parse_url
from ..utils import url_assembly
from ..utils import dict_full_update
from .. import common
from ..logger import logger
import inspect

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()


class CustomNotifyPlugin(NotifyBase):
    """
    Apprise Custom Plugin Hook

    This gets initialized based on @notify decorator definitions

    """
    # Our Custom notification
    service_url = 'https://github.com/caronc/apprise/wiki/Custom_Notification'

    # Over-ride our category since this inheritance of the NotifyBase class
    # should be treated differently.
    category = 'custom'

    # Define object templates
    templates = (
        '{schema}://',
    )

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns arguments retrieved

        """
        return parse_url(url, verify_host=False, simple=True)

    def url(self, privacy=False, *args, **kwargs):
        """
        General URL assembly
        """
        return '{schema}://'.format(schema=self.secure_protocol)

    @staticmethod
    def instantiate_plugin(url, send_func, name=None):
        """
        The function used to add a new notification plugin based on the schema
        parsed from the provided URL into our supported matrix structure.
        """

        if not isinstance(url, str):
            msg = 'An invalid custom notify url/schema ({}) provided in ' \
                'function {}.'.format(url, send_func.__name__)
            logger.warning(msg)
            return None

        # Validate that our schema is okay
        re_match = URL_DETAILS_RE.match(url)
        if not re_match:
            msg = 'An invalid custom notify url/schema ({}) provided in ' \
                'function {}.'.format(url, send_func.__name__)
            logger.warning(msg)
            return None

        # Acquire our schema
        schema = re_match.group('schema').lower()

        if not re_match.group('base'):
            url = '{}://'.format(schema)

        # Keep a default set of arguments to apply to all called references
        base_args = parse_url(
            url, default_schema=schema, verify_host=False, simple=True)

        if schema in N_MGR:
            # we're already handling this object
            msg = 'The schema ({}) is already defined and could not be ' \
                'loaded from custom notify function {}.' \
                .format(url, send_func.__name__)
            logger.warning(msg)
            return None

        # We define our own custom wrapper class so that we can initialize
        # some key default configuration values allowing calls to our
        # `Apprise.details()` to correctly differentiate one custom plugin
        # that was loaded from another
        class CustomNotifyPluginWrapper(CustomNotifyPlugin):

            # Our Service Name
            service_name = name if isinstance(name, str) \
                and name else 'Custom - {}'.format(schema)

            # Store our matched schema
            secure_protocol = schema

            requirements = {
                # Define our required packaging in order to work
                'details': "Source: {}".format(inspect.getfile(send_func))
            }

            # Assign our send() function
            __send = staticmethod(send_func)

            # Update our default arguments
            _base_args = base_args

            def __init__(self, **kwargs):
                """
                Our initialization

                """
                #  init parent
                super().__init__(**kwargs)

                self._default_args = {}

                # Apply our updates based on what was parsed
                dict_full_update(self._default_args, self._base_args)
                dict_full_update(self._default_args, kwargs)

                # Update our arguments (applying them to what we originally)
                # initialized as
                self._default_args['url'] = url_assembly(**self._default_args)

            def send(self, body, title='', notify_type=common.NotifyType.INFO,
                     *args, **kwargs):
                """
                Our send() call which triggers our hook
                """

                response = False
                try:
                    # Enforce a boolean response
                    result = self.__send(
                        body, title, notify_type, *args,
                        meta=self._default_args, **kwargs)

                    if result is None:
                        # The wrapper did not define a return (or returned
                        # None)
                        # this is treated as a successful return as it is
                        # assumed the developer did not care about the result
                        # of the call.
                        response = True

                    else:
                        # Perform boolean check (allowing obects to also be
                        # returned and check against the __bool__ call
                        response = True if result else False

                except Exception as e:
                    # Unhandled Exception
                    self.logger.warning(
                        'An exception occured sending a %s notification.',
                        N_MGR[self.secure_protocol].service_name)
                    self.logger.debug(
                        '%s Exception: %s',
                        N_MGR[self.secure_protocol], str(e))
                    return False

                if response:
                    self.logger.info(
                        'Sent %s notification.',
                        N_MGR[self.secure_protocol].service_name)
                else:
                    self.logger.warning(
                        'Failed to send %s notification.',
                        N_MGR[self.secure_protocol].service_name)
                return response

        # Store our plugin into our core map file
        return N_MGR.add(
            plugin=CustomNotifyPluginWrapper,
            schemas=schema,
            send_func=send_func,
            url=url,
        )



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from unittest import mock

import pytest
import requests

from apprise.plugins.NotifyFlock import NotifyFlock
from helpers import AppriseURLTester

# Disable logging for a cleaner testing output
import logging
logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    ('flock://', {
        'instance': TypeError,
    }),
    # An invalid url
    ('flock://:@/', {
        'instance': TypeError,
    }),
    # Provide a token
    ('flock://%s' % ('t' * 24), {
        'instance': NotifyFlock,
    }),
    # Image handling
    ('flock://%s?image=True' % ('t' * 24), {
        'instance': NotifyFlock,

        # Our expected url(privacy=True) startswith() response:
        'privacy_url': 'flock://t...t',
    }),
    ('flock://%s?image=False' % ('t' * 24), {
        'instance': NotifyFlock,
    }),
    ('flock://%s?image=True' % ('t' * 24), {
        'instance': NotifyFlock,
        # Run test when image is set to True, but one couldn't actually be
        # loaded from the Asset Object.
        'include_image': False,
    }),
    # Test to=
    ('flock://%s?to=u:%s&format=markdown' % ('i' * 24, 'u' * 12), {
        'instance': NotifyFlock,
    }),
    # Provide markdown format
    ('flock://%s?format=markdown' % ('i' * 24), {
        'instance': NotifyFlock,
    }),
    # Provide text format
    ('flock://%s?format=text' % ('i' * 24), {
        'instance': NotifyFlock,
    }),
    # Native URL Support, take the slack URL and still build from it
    ('https://api.flock.com/hooks/sendMessage/{}/'.format('i' * 24), {
        'instance': NotifyFlock,
    }),
    # Native URL Support with arguments
    ('https://api.flock.com/hooks/sendMessage/{}/?format=markdown'.format(
        'i' * 24), {
        'instance': NotifyFlock,
    }),
    # Bot API presumed if one or more targets are specified
    # Provide markdown format
    ('flock://%s/u:%s?format=markdown' % ('i' * 24, 'u' * 12), {
        'instance': NotifyFlock,
    }),
    # Bot API presumed if one or more targets are specified
    # Provide text format
    ('flock://%s/u:%s?format=html' % ('i' * 24, 'u' * 12), {
        'instance': NotifyFlock,
    }),
    # Bot API presumed if one or more targets are specified
    # u: is optional
    ('flock://%s/%s?format=text' % ('i' * 24, 'u' * 12), {
        'instance': NotifyFlock,
    }),
    # Bot API presumed if one or more targets are specified
    # Multi-entries
    ('flock://%s/g:%s/u:%s?format=text' % ('i' * 24, 'g' * 12, 'u' * 12), {
        'instance': NotifyFlock,
    }),
    # Bot API presumed if one or more targets are specified
    # Multi-entries using @ for user and # for channel
    ('flock://%s/#%s/@%s?format=text' % ('i' * 24, 'g' * 12, 'u' * 12), {
        'instance': NotifyFlock,
    }),
    # Bot API presumed if one or more targets are specified
    # has bad entry
    ('flock://%s/g:%s/u:%s?format=text' % ('i' * 24, 'g' * 12, 'u' * 10), {
        'instance': NotifyFlock,
    }),
    # Invalid user/group defined
    ('flock://%s/g:/u:?format=text' % ('i' * 24), {
        'instance': TypeError,
    }),
    # we don't focus on the invalid length of the user/group fields.
    # As a result, the following will load and pass the data upstream
    ('flock://%s/g:%s/u:%s?format=text' % ('i' * 24, 'g' * 14, 'u' * 10), {
        # We will still instantiate the object
        'instance': NotifyFlock,
    }),
    # Error Testing
    ('flock://%s/g:%s/u:%s?format=text' % ('i' * 24, 'g' * 12, 'u' * 10), {
        'instance': NotifyFlock,
        # force a failure
        'response': False,
        'requests_response_code': requests.codes.internal_server_error,
    }),
    ('flock://%s/' % ('t' * 24), {
        'instance': NotifyFlock,
        # force a failure
        'response': False,
        'requests_response_code': requests.codes.internal_server_error,
    }),
    ('flock://%s/' % ('t' * 24), {
        'instance': NotifyFlock,
        # throw a bizzare code forcing us to fail to look it up
        'response': False,
        'requests_response_code': 999,
    }),
    ('flock://%s/' % ('t' * 24), {
        'instance': NotifyFlock,
        # Throws a series of connection and transfer exceptions when this flag
        # is set and tests that we gracfully handle them
        'test_requests_exceptions': True,
    }),
)


def test_plugin_flock_urls():
    """
    NotifyFlock() Apprise URLs

    """

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch('requests.get')
@mock.patch('requests.post')
def test_plugin_flock_edge_cases(mock_post, mock_get):
    """
    NotifyFlock() Edge Cases

    """

    # Initializes the plugin with an invalid token
    with pytest.raises(TypeError):
        NotifyFlock(token=None)
    # Whitespace also acts as an invalid token value
    with pytest.raises(TypeError):
        NotifyFlock(token="   ")



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from unittest import mock

import requests
import pytest
from json import dumps
from apprise.plugins.NotifyPushover import PushoverPriority, NotifyPushover
import apprise
from helpers import AppriseURLTester

# Disable logging for a cleaner testing output
import logging
logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), 'var')

# Our Testing URLs
apprise_url_tests = (
    ('pover://', {
        'instance': TypeError,
    }),
    # bad url
    ('pover://:@/', {
        'instance': TypeError,
    }),
    # APIkey; no user
    ('pover://%s' % ('a' * 30), {
        'instance': TypeError,
    }),
    # API Key + custom sound setting
    ('pover://%s@%s?sound=mysound' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + valid alternate sound picked
    ('pover://%s@%s?sound=spacealarm' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + valid url_title with url
    ('pover://%s@%s?url=my-url&url_title=title' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + Valid User
    ('pover://%s@%s' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
        # don't include an image by default
        'include_image': False,
    }),
    # API Key + Valid User + 1 Device
    ('pover://%s@%s/DEVICE' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + Valid User + 1 Device (via to=)
    ('pover://%s@%s?to=DEVICE' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + Valid User + 2 Devices
    ('pover://%s@%s/DEVICE1/Device-with-dash/' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,

        # Our expected url(privacy=True) startswith() response:
        'privacy_url': 'pover://u...u@a...a',
    }),
    # API Key + Valid User + invalid device
    ('pover://%s@%s/%s/' % ('u' * 30, 'a' * 30, 'd' * 30), {
        'instance': NotifyPushover,
        # Notify will return False since there is a bad device in our list
        'response': False,
    }),
    # API Key + Valid User + device + invalid device
    ('pover://%s@%s/DEVICE1/%s/' % ('u' * 30, 'a' * 30, 'd' * 30), {
        'instance': NotifyPushover,
        # Notify will return False since there is a bad device in our list
        'response': False,
    }),
    # API Key + priority setting
    ('pover://%s@%s?priority=high' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + priority setting + html mode
    ('pover://%s@%s?priority=high&format=html' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + priority setting + markdown mode
    ('pover://%s@%s?priority=high&format=markdown' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + invalid priority setting
    ('pover://%s@%s?priority=invalid' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + emergency(2) priority setting
    ('pover://%s@%s?priority=emergency' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + emergency(2) priority setting (via numeric value
    ('pover://%s@%s?priority=2' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    # API Key + emergency priority setting with retry and expire
    ('pover://%s@%s?priority=emergency&%s&%s' % ('u' * 30,
                                                 'a' * 30,
                                                 'retry=30',
                                                 'expire=300'), {
        'instance': NotifyPushover,
    }),
    # API Key + emergency priority setting with text retry
    ('pover://%s@%s?priority=emergency&%s&%s' % ('u' * 30,
                                                 'a' * 30,
                                                 'retry=invalid',
                                                 'expire=300'), {
        'instance': NotifyPushover,
    }),
    # API Key + emergency priority setting with text expire
    ('pover://%s@%s?priority=emergency&%s&%s' % ('u' * 30,
                                                 'a' * 30,
                                                 'retry=30',
                                                 'expire=invalid'), {
        'instance': NotifyPushover,
    }),
    # API Key + emergency priority setting with invalid expire
    ('pover://%s@%s?priority=emergency&%s' % ('u' * 30,
                                              'a' * 30,
                                              'expire=100000'), {
        'instance': TypeError,
    }),
    # API Key + emergency priority setting with invalid retry
    ('pover://%s@%s?priority=emergency&%s' % ('u' * 30,
                                              'a' * 30,
                                              'retry=15'), {
        'instance': TypeError,
    }),
    # API Key + priority setting (empty)
    ('pover://%s@%s?priority=' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
    }),
    ('pover://%s@%s' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
        # force a failure
        'response': False,
        'requests_response_code': requests.codes.internal_server_error,
    }),
    ('pover://%s@%s' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
        # throw a bizzare code forcing us to fail to look it up
        'response': False,
        'requests_response_code': 999,
    }),
    ('pover://%s@%s' % ('u' * 30, 'a' * 30), {
        'instance': NotifyPushover,
        # Throws a series of connection and transfer exceptions when this flag
        # is set and tests that we gracfully handle them
        'test_requests_exceptions': True,
    }),
)


def test_plugin_pushover_urls():
    """
    NotifyPushover() Apprise URLs

    """

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch('requests.post')
def test_plugin_pushover_attachments(mock_post, tmpdir):
    """
    NotifyPushover() Attachment Checks

    """

    # Initialize some generic (but valid) tokens
    user_key = 'u' * 30
    api_token = 'a' * 30

    # Prepare a good response
    response = mock.Mock()
    response.content = dumps(
        {"status": 1, "request": "647d2300-702c-4b38-8b2f-d56326ae460b"})
    response.status_code = requests.codes.ok

    # Prepare a bad response
    bad_response = mock.Mock()
    response.content = dumps(
        {"status": 1, "request": "647d2300-702c-4b38-8b2f-d56326ae460b"})
    bad_response.status_code = requests.codes.internal_server_error

    # Assign our good response
    mock_post.return_value = response

    # prepare our attachment
    attach = apprise.AppriseAttachment(
        os.path.join(TEST_VAR_DIR, 'apprise-test.gif'))

    # Instantiate our object
    obj = apprise.Apprise.instantiate(
        'pover://{}@{}/'.format(user_key, api_token))
    assert isinstance(obj, NotifyPushover)

    # Test our attachment
    assert obj.notify(body="test", attach=attach) is True

    # Test our call count
    assert mock_post.call_count == 1
    assert mock_post.call_args_list[0][0][0] == \
        'https://api.pushover.net/1/messages.json'

    # Reset our mock object for multiple tests
    mock_post.reset_mock()

    # Test multiple attachments
    assert attach.add(os.path.join(TEST_VAR_DIR, 'apprise-test.gif'))
    assert obj.notify(body="test", attach=attach) is True

    # Test our call count
    assert mock_post.call_count == 2
    assert mock_post.call_args_list[0][0][0] == \
        'https://api.pushover.net/1/messages.json'
    assert mock_post.call_args_list[1][0][0] == \
        'https://api.pushover.net/1/messages.json'

    # Reset our mock object for multiple tests
    mock_post.reset_mock()

    image = tmpdir.mkdir("pover_image").join("test.jpg")
    image.write('a' * NotifyPushover.attach_max_size_bytes)

    attach = apprise.AppriseAttachment.instantiate(str(image))
    assert obj.notify(body="test", attach=attach) is True

    # Test our call count
    assert mock_post.call_count == 1
    assert mock_post.call_args_list[0][0][0] == \
        'https://api.pushover.net/1/messages.json'

    # Reset our mock object for multiple tests
    mock_post.reset_mock()

    # Add 1 more byte to the file (putting it over the limit)
    image.write(
        'a' * (NotifyPushover.attach_max_size_bytes + 1))

    attach = apprise.AppriseAttachment.instantiate(str(image))
    assert obj.notify(body="test", attach=attach) is False

    # Test our call count
    assert mock_post.call_count == 0

    # Test case when file is missing
    attach = apprise.AppriseAttachment.instantiate(
        'file://{}?cache=False'.format(str(image)))
    os.unlink(str(image))
    assert obj.notify(
        body='body', title='title', attach=attach) is False

    # Test our call count
    assert mock_post.call_count == 0

    # Test unsuported files:
    image = tmpdir.mkdir("pover_unsupported").join("test.doc")
    image.write('a' * 256)
    attach = apprise.AppriseAttachment.instantiate(str(image))

    # Content is silently ignored
    assert obj.notify(body="test", attach=attach) is True

    # prepare our attachment
    attach = apprise.AppriseAttachment(
        os.path.join(TEST_VAR_DIR, 'apprise-test.gif'))

    # Throw an exception on the first call to requests.post()
    for side_effect in (requests.RequestException(), OSError(), bad_response):
        mock_post.side_effect = [side_effect, side_effect]

        # We'll fail now because of our error handling
        assert obj.send(body="test", attach=attach) is False

        # Same case without an attachment
        assert obj.send(body="test") is False


@mock.patch('requests.post')
def test_plugin_pushover_edge_cases(mock_post):
    """
    NotifyPushover() Edge Cases

    """

    # No token
    with pytest.raises(TypeError):
        NotifyPushover(token=None)

    # Initialize some generic (but valid) tokens
    token = 'a' * 30
    user_key = 'u' * 30

    invalid_device = 'd' * 35

    # Support strings
    devices = 'device1,device2,,,,%s' % invalid_device

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    # No webhook id specified
    with pytest.raises(TypeError):
        NotifyPushover(user_key=user_key, webhook_id=None)

    obj = NotifyPushover(
        user_key=user_key, token=token, targets=devices)
    assert isinstance(obj, NotifyPushover) is True
    # Our invalid device is ignored
    assert len(obj.targets) == 2

    # We notify the 2 devices loaded
    assert obj.notify(
        body='body', title='title',
        notify_type=apprise.NotifyType.INFO) is True

    obj = NotifyPushover(user_key=user_key, token=token)
    assert isinstance(obj, NotifyPushover) is True
    # Default is to send to all devices, so there will be a
    # device defined here
    assert len(obj.targets) == 1

    # This call succeeds because all of the devices are valid
    assert obj.notify(
        body='body', title='title',
        notify_type=apprise.NotifyType.INFO) is True

    obj = NotifyPushover(
        user_key=user_key, token=token, targets=set())
    assert isinstance(obj, NotifyPushover) is True
    # Default is to send to all devices, so there will be a
    # device defined here
    assert len(obj.targets) == 1

    # No User Key specified
    with pytest.raises(TypeError):
        NotifyPushover(user_key=None, token="abcd")

    # No Access Token specified
    with pytest.raises(TypeError):
        NotifyPushover(user_key="abcd", token=None)

    with pytest.raises(TypeError):
        NotifyPushover(user_key="abcd", token="  ")


@mock.patch('requests.post')
def test_plugin_pushover_config_files(mock_post):
    """
    NotifyPushover() Config File Cases
    """
    content = """
    urls:
      - pover://USER@TOKEN:
          - priority: -2
            tag: pushover_int low
          - priority: "-2"
            tag: pushover_str_int low
          - priority: low
            tag: pushover_str low

          # This will take on normal (default) priority
          - priority: invalid
            tag: pushover_invalid

      - pover://USER2@TOKEN2:
          - priority: 2
            tag: pushover_int emerg
          - priority: "2"
            tag: pushover_str_int emerg
          - priority: emergency
            tag: pushover_str emerg
    """

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    # Create ourselves a config object
    ac = apprise.AppriseConfig()
    assert ac.add_config(content=content) is True

    aobj = apprise.Apprise()

    # Add our configuration
    aobj.add(ac)

    # We should be able to read our 7 servers from that
    # 3x low
    # 3x emerg
    # 1x invalid (so takes on normal priority)
    assert len(ac.servers()) == 7
    assert len(aobj) == 7
    assert len([x for x in aobj.find(tag='low')]) == 3
    for s in aobj.find(tag='low'):
        assert s.priority == PushoverPriority.LOW

    assert len([x for x in aobj.find(tag='emerg')]) == 3
    for s in aobj.find(tag='emerg'):
        assert s.priority == PushoverPriority.EMERGENCY

    assert len([x for x in aobj.find(tag='pushover_str')]) == 2
    assert len([x for x in aobj.find(tag='pushover_str_int')]) == 2
    assert len([x for x in aobj.find(tag='pushover_int')]) == 2

    assert len([x for x in aobj.find(tag='pushover_invalid')]) == 1
    assert next(aobj.find(tag='pushover_invalid')).priority == \
        PushoverPriority.NORMAL

    # Notifications work
    # We test 'pushover_str_int' and 'low' which only matches 1 end point
    assert aobj.notify(
        title="title", body="body", tag=[('pushover_str_int', 'low')]) is True

    # Notify everything loaded
    assert aobj.notify(title="title", body="body") is True



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# For this plugin to work, you need to add the Maker applet to your profile
# Simply visit https://ifttt.com/search and search for 'Webhooks'
# Or if you're signed in, click here: https://ifttt.com/maker_webhooks
# and click 'Connect'
#
# You'll want to visit the settings of this Applet and pay attention to the
# URL. For example, it might look like this:
#               https://maker.ifttt.com/use/a3nHB7gA9TfBQSqJAHklod
#
# In the above example a3nHB7gA9TfBQSqJAHklod becomes your {webhook_id}
# You will need this to make this notification work correctly
#
# For each event you create you will assign it a name (this will be known as
# the {event} when building your URL.
import re
import requests
from json import dumps

from .NotifyBase import NotifyBase
from ..common import NotifyType
from ..utils import parse_list
from ..utils import validate_regex
from ..AppriseLocale import gettext_lazy as _


class NotifyIFTTT(NotifyBase):
    """
    A wrapper for IFTTT Notifications

    """

    # The default descriptive name associated with the Notification
    service_name = 'IFTTT'

    # The services URL
    service_url = 'https://ifttt.com/'

    # The default protocol
    secure_protocol = 'ifttt'

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_ifttt'

    # Even though you'll add 'Ingredients' as {{ Value1 }} to your Applets,
    # you must use their lowercase value in the HTTP POST.
    ifttt_default_key_prefix = 'value'

    # The default IFTTT Key to use when mapping the title text to the IFTTT
    # event. The idea here is if someone wants to over-ride the default and
    # change it to another Ingredient Name (in 2018, you were limited to have
    # value1, value2, and value3).
    ifttt_default_title_key = 'value1'

    # The default IFTTT Key to use when mapping the body text to the IFTTT
    # event. The idea here is if someone wants to over-ride the default and
    # change it to another Ingredient Name (in 2018, you were limited to have
    # value1, value2, and value3).
    ifttt_default_body_key = 'value2'

    # The default IFTTT Key to use when mapping the body text to the IFTTT
    # event. The idea here is if someone wants to over-ride the default and
    # change it to another Ingredient Name (in 2018, you were limited to have
    # value1, value2, and value3).
    ifttt_default_type_key = 'value3'

    # IFTTT uses the http protocol with JSON requests
    notify_url = 'https://maker.ifttt.com/' \
                 'trigger/{event}/with/key/{webhook_id}'

    # Define object templates
    templates = (
        '{schema}://{webhook_id}/{events}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'webhook_id': {
            'name': _('Webhook ID'),
            'type': 'string',
            'private': True,
            'required': True,
        },
        'events': {
            'name': _('Events'),
            'type': 'list:string',
            'required': True,
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'to': {
            'alias_of': 'events',
        },
    })

    # Define our token control
    template_kwargs = {
        'add_tokens': {
            'name': _('Add Tokens'),
            'prefix': '+',
        },
        'del_tokens': {
            'name': _('Remove Tokens'),
            'prefix': '-',
        },
    }

    def __init__(self, webhook_id, events, add_tokens=None, del_tokens=None,
                 **kwargs):
        """
        Initialize IFTTT Object

        add_tokens can optionally be a dictionary of key/value pairs
        that you want to include in the IFTTT post to the server.

        del_tokens can optionally be a list/tuple/set of tokens
        that you want to eliminate from the IFTTT post.  There isn't
        much real functionality to this one unless you want to remove
        reference to Value1, Value2, and/or Value3

        """
        super().__init__(**kwargs)

        # Webhook ID (associated with project)
        self.webhook_id = validate_regex(webhook_id)
        if not self.webhook_id:
            msg = 'An invalid IFTTT Webhook ID ' \
                  '({}) was specified.'.format(webhook_id)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our Events we wish to trigger
        self.events = parse_list(events)
        if not self.events:
            msg = 'You must specify at least one event you wish to trigger on.'
            self.logger.warning(msg)
            raise TypeError(msg)

        # Tokens to include in post
        self.add_tokens = {}
        if add_tokens:
            self.add_tokens.update(add_tokens)

        # Tokens to remove
        self.del_tokens = []
        if del_tokens is not None:
            if isinstance(del_tokens, (list, tuple, set)):
                self.del_tokens = del_tokens

            elif isinstance(del_tokens, dict):
                # Convert the dictionary into a list
                self.del_tokens = set(del_tokens.keys())

            else:
                msg = 'del_token must be a list; {} was provided'.format(
                    str(type(del_tokens)))
                self.logger.warning(msg)
                raise TypeError(msg)

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        Perform IFTTT Notification
        """

        headers = {
            'User-Agent': self.app_id,
            'Content-Type': 'application/json',
        }

        # prepare JSON Object
        payload = {
            self.ifttt_default_title_key: title,
            self.ifttt_default_body_key: body,
            self.ifttt_default_type_key: notify_type,
        }

        # Add any new tokens expected (this can also potentially override
        # any entries defined above)
        payload.update(self.add_tokens)

        # Eliminate fields flagged for removal otherwise ensure all tokens are
        # lowercase since that is what the IFTTT server expects from us.
        payload = {x.lower(): y for x, y in payload.items()
                   if x not in self.del_tokens}

        # error tracking (used for function return)
        has_error = False

        # Create a copy of our event lit
        events = list(self.events)

        while len(events):

            # Retrive an entry off of our event list
            event = events.pop(0)

            # URL to transmit content via
            url = self.notify_url.format(
                webhook_id=self.webhook_id,
                event=event,
            )

            self.logger.debug('IFTTT POST URL: %s (cert_verify=%r)' % (
                url, self.verify_certificate,
            ))
            self.logger.debug('IFTTT Payload: %s' % str(payload))

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                self.logger.debug(
                    u"IFTTT HTTP response headers: %r" % r.headers)
                self.logger.debug(
                    u"IFTTT HTTP response body: %r" % r.content)

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = \
                        NotifyIFTTT.http_response_code_lookup(r.status_code)

                    self.logger.warning(
                        'Failed to send IFTTT notification to {}: '
                        '{}{}error={}.'.format(
                            event,
                            status_str,
                            ', ' if status_str else '',
                            r.status_code))

                    self.logger.debug(
                        'Response Details:\r\n{}'.format(r.content))

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        'Sent IFTTT notification to %s.' % event)

            except requests.RequestException as e:
                self.logger.warning(
                    'A Connection error occurred sending IFTTT:%s ' % (
                        event) + 'notification.'
                )
                self.logger.debug('Socket Exception: %s' % str(e))

                # Mark our failure
                has_error = True
                continue

        return not has_error

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        # Store any new key/value pairs added to our list
        params.update({'+{}'.format(k): v for k, v in self.add_tokens})
        params.update({'-{}'.format(k): '' for k in self.del_tokens})

        return '{schema}://{webhook_id}@{events}/?{params}'.format(
            schema=self.secure_protocol,
            webhook_id=self.pprint(self.webhook_id, privacy, safe=''),
            events='/'.join([NotifyIFTTT.quote(x, safe='')
                             for x in self.events]),
            params=NotifyIFTTT.urlencode(params),
        )

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        return len(self.events)

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Our API Key is the hostname if no user is specified
        results['webhook_id'] = \
            results['user'] if results['user'] else results['host']

        # Unquote our API Key
        results['webhook_id'] = NotifyIFTTT.unquote(results['webhook_id'])

        # Parse our add_token and del_token arguments (if specified)
        results['add_token'] = results['qsd+']
        results['del_token'] = results['qsd-']

        # Our Event
        results['events'] = list()
        if results['user']:
            # If a user was defined, then the hostname is actually a event
            # too
            results['events'].append(NotifyIFTTT.unquote(results['host']))

        # Now fetch the remaining tokens
        results['events'].extend(NotifyIFTTT.split_path(results['fullpath']))

        # The 'to' makes it easier to use yaml configuration
        if 'to' in results['qsd'] and len(results['qsd']['to']):
            results['events'] += \
                NotifyIFTTT.parse_list(results['qsd']['to'])

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://maker.ifttt.com/use/WEBHOOK_ID/EVENT_ID
        """

        result = re.match(
            r'^https?://maker\.ifttt\.com/use/'
            r'(?P<webhook_id>[A-Z0-9_-]+)'
            r'((?P<events>(/[A-Z0-9_-]+)+))?'
            r'/?(?P<params>\?.+)?$', url, re.I)

        if result:
            return NotifyIFTTT.parse_url(
                '{schema}://{webhook_id}{events}{params}'.format(
                    schema=NotifyIFTTT.secure_protocol,
                    webhook_id=result.group('webhook_id'),
                    events='' if not result.group('events')
                    else '@{}'.format(result.group('events')),
                    params='' if not result.group('params')
                    else result.group('params')))

        return None
